from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import asset


class InitializeMarketArgsJSON(typing.TypedDict):
    asset: asset.AssetJSON
    market_nonce: int
    base_mint_nonce: int
    quote_mint_nonce: int
    zeta_base_vault_nonce: int
    zeta_quote_vault_nonce: int
    dex_base_vault_nonce: int
    dex_quote_vault_nonce: int
    vault_signer_nonce: int


@dataclass
class InitializeMarketArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "asset" / asset.layout,
        "market_nonce" / borsh.U8,
        "base_mint_nonce" / borsh.U8,
        "quote_mint_nonce" / borsh.U8,
        "zeta_base_vault_nonce" / borsh.U8,
        "zeta_quote_vault_nonce" / borsh.U8,
        "dex_base_vault_nonce" / borsh.U8,
        "dex_quote_vault_nonce" / borsh.U8,
        "vault_signer_nonce" / borsh.U64,
    )
    asset: asset.AssetKind
    market_nonce: int
    base_mint_nonce: int
    quote_mint_nonce: int
    zeta_base_vault_nonce: int
    zeta_quote_vault_nonce: int
    dex_base_vault_nonce: int
    dex_quote_vault_nonce: int
    vault_signer_nonce: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeMarketArgs":
        return cls(
            asset=asset.from_decoded(obj.asset),
            market_nonce=obj.market_nonce,
            base_mint_nonce=obj.base_mint_nonce,
            quote_mint_nonce=obj.quote_mint_nonce,
            zeta_base_vault_nonce=obj.zeta_base_vault_nonce,
            zeta_quote_vault_nonce=obj.zeta_quote_vault_nonce,
            dex_base_vault_nonce=obj.dex_base_vault_nonce,
            dex_quote_vault_nonce=obj.dex_quote_vault_nonce,
            vault_signer_nonce=obj.vault_signer_nonce,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "asset": self.asset.to_encodable(),
            "market_nonce": self.market_nonce,
            "base_mint_nonce": self.base_mint_nonce,
            "quote_mint_nonce": self.quote_mint_nonce,
            "zeta_base_vault_nonce": self.zeta_base_vault_nonce,
            "zeta_quote_vault_nonce": self.zeta_quote_vault_nonce,
            "dex_base_vault_nonce": self.dex_base_vault_nonce,
            "dex_quote_vault_nonce": self.dex_quote_vault_nonce,
            "vault_signer_nonce": self.vault_signer_nonce,
        }

    def to_json(self) -> InitializeMarketArgsJSON:
        return {
            "asset": self.asset.to_json(),
            "market_nonce": self.market_nonce,
            "base_mint_nonce": self.base_mint_nonce,
            "quote_mint_nonce": self.quote_mint_nonce,
            "zeta_base_vault_nonce": self.zeta_base_vault_nonce,
            "zeta_quote_vault_nonce": self.zeta_quote_vault_nonce,
            "dex_base_vault_nonce": self.dex_base_vault_nonce,
            "dex_quote_vault_nonce": self.dex_quote_vault_nonce,
            "vault_signer_nonce": self.vault_signer_nonce,
        }

    @classmethod
    def from_json(cls, obj: InitializeMarketArgsJSON) -> "InitializeMarketArgs":
        return cls(
            asset=asset.from_json(obj["asset"]),
            market_nonce=obj["market_nonce"],
            base_mint_nonce=obj["base_mint_nonce"],
            quote_mint_nonce=obj["quote_mint_nonce"],
            zeta_base_vault_nonce=obj["zeta_base_vault_nonce"],
            zeta_quote_vault_nonce=obj["zeta_quote_vault_nonce"],
            dex_base_vault_nonce=obj["dex_base_vault_nonce"],
            dex_quote_vault_nonce=obj["dex_quote_vault_nonce"],
            vault_signer_nonce=obj["vault_signer_nonce"],
        )



from solana.rpc import commitment
from solders.address_lookup_table_account import AddressLookupTableAccount
from solders.pubkey import Pubkey

from zetamarkets_py.types import Asset, Network

ZETA_PID = {
    # Network.LOCALNET: Pubkey.from_string("BG3oRikW8d16YjUEmX3ZxHm9SiJzrGtMhsSR8aCw1Cd7"),
    Network.DEVNET: Pubkey.from_string("BG3oRikW8d16YjUEmX3ZxHm9SiJzrGtMhsSR8aCw1Cd7"),
    Network.MAINNET: Pubkey.from_string("ZETAxsqBRek56DhiGXrn75yj2NHU3aYUnxvHXpkf3aD"),
}

MATCHING_ENGINE_PID = {
    # Network.LOCALNET: Pubkey.from_string("5CmWtUihvSrJpaUrpJ3H1jUa9DRjYz4v2xs6c3EgQWMf"),
    Network.DEVNET: Pubkey.from_string("5CmWtUihvSrJpaUrpJ3H1jUa9DRjYz4v2xs6c3EgQWMf"),
    Network.MAINNET: Pubkey.from_string("zDEXqXEG7gAyxb1Kg9mK5fPnUdENCGKzWrM21RMdWRq"),
}

# Asset keys are wormhole from mainnet.
MINTS = {
    Asset.SOL: Pubkey.from_string("So11111111111111111111111111111111111111112"),
    Asset.BTC: Pubkey.from_string("qfnqNqs3nCAHjnyCgLRDbBtq4p2MtHZxw8YjSyYhPoL"),
    Asset.ETH: Pubkey.from_string("FeGn77dhg1KXRRFeSwwMiykZnZPw5JXW6naf2aQgZDQf"),
}

# These are generated flexible PDAs and aren't reflective of an spl token mint.
FLEXIBLE_MINTS = {
    Network.DEVNET: {
        Asset.APT: Pubkey.from_string("FbfkphUHaAd7c27RqhzKBRAPX8T5AzFBH259sbGmNuvG"),
        Asset.ARB: Pubkey.from_string("w8h6r5ogLihfuWeCA1gs7boxNjzbwWeQbXMB3UATaC6"),
    },
    Network.MAINNET: {
        Asset.APT: Pubkey.from_string("8z8oShLky1PauW9hxv6AsjnricLqoK9MfmNZJDQNNNPr"),
        Asset.ARB: Pubkey.from_string("Ebd7aUFu3rtsZruCzTnG4tjBoxaJdWT8S3t4yC8hVpbo"),
    },
}

USDC_MINT = {
    #   Network.LOCALNET: Pubkey.from_string("6PEh8n3p7BbCTykufbq1nSJYAZvUp6gSwEANAs1ZhsCX"),
    Network.DEVNET: Pubkey.from_string("6PEh8n3p7BbCTykufbq1nSJYAZvUp6gSwEANAs1ZhsCX"),
    Network.MAINNET: Pubkey.from_string("EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v"),
}

CHAINLINK_PID = Pubkey.from_string("HEvSKofvBgfaexv23kMabbYqxasxU3mQ4ibBMEmJWHny")

MAX_SETTLE_AND_CLOSE_PER_TX = 4
MAX_CANCELS_PER_TX = 3
MAX_CANCELS_PER_TX_LUT = 13
MAX_GREEK_UPDATES_PER_TX = 20
MAX_SETTLEMENT_ACCOUNTS = 20
MAX_FUNDING_ACCOUNTS = 20
MAX_REBALANCE_ACCOUNTS = 18
MAX_SETTLE_ACCOUNTS = 5
MAX_ZETA_GROUPS = 20
MAX_MARGIN_AND_SPREAD_ACCOUNTS = 20
MAX_SET_REFERRALS_REWARDS_ACCOUNTS = 12
MAX_INITIALIZE_MARKET_TIF_EPOCH_CYCLE_IXS_PER_TX = 15
MARKET_INDEX_LIMIT = 18
MARKET_LOAD_LIMIT = 12
MAX_MARKETS_TO_FETCH = 50

MIN_LOT_SIZE = 0.001
PERP_MARKET_ORDER_SPOT_SLIPPAGE = 0.02

# This is the most we can load per iteration without
# hitting the rate limit.
MARKET_LOAD_LIMIT = 12

# Numbers represented in BN are generally fixed point integers with precision of 6.
PLATFORM_PRECISION = 6
MARGIN_PRECISION = 8
POSITION_PRECISION = 3
TICK_SIZE = 100

DEFAULT_ORDER_TAG = "SDK"

MAX_POSITION_MOVEMENTS = 10
BPS_DENOMINATOR = 10_000

DEFAULT_MICRO_LAMPORTS_PER_CU_FEE = 1000

# DEX
BASE_MINT_DECIMALS = 0
QUOTE_MINT_DECIMALS = 6

BLOCKHASH_COMMITMENT = commitment.Finalized

ZETA_LUT = {
    # Network.LOCALNET: None,
    Network.DEVNET: AddressLookupTableAccount(
        key=Pubkey.from_string("3a2We1NeafvL9Hnx41ZXZQRjxLxr8M6MnjneZZjJDRfZ"),
        addresses=[
            Pubkey.from_string("9VddCF6iEyZbjkCQ4g8VJpjEtuLsgmvRCc6LQwAvXigC"),
            Pubkey.from_string("BditorsEXbVw6R8Woe6JhoyXC1beJCixYt57UMKuiwQi"),
            Pubkey.from_string("5CmWtUihvSrJpaUrpJ3H1jUa9DRjYz4v2xs6c3EgQWMf"),
            Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"),
            Pubkey.from_string("7m134nU79ezr3b7jYSehnhto94AvSP4yexa3Wdhxff99"),
            Pubkey.from_string("SysvarRent111111111111111111111111111111111"),
            Pubkey.from_string("HEvSKofvBgfaexv23kMabbYqxasxU3mQ4ibBMEmJWHny"),
            Pubkey.from_string("J83w4HKfqxwcq3BEMMkPFSppX3gqekLyLJBexebFVkix"),
            Pubkey.from_string("99B2bTijsU6f1GCT73HmdR7HCFFjGMBcPZY6jZ96ynrR"),
            Pubkey.from_string("HovQMDrbAgAYPCmHVSrezcSmkMtXSSUsLDFANExrZh2J"),
            Pubkey.from_string("6PxBx93S8x3tno1TsFZwT5VqP8drrRCbCXygEXYNkFJe"),
            Pubkey.from_string("EdVCmQ9FSPcVe5YySXDPCRmc8aDQLKJ9xvYBMZPie1Vw"),
            Pubkey.from_string("669U43LNHx7LsVj95uYksnhXUfWKDsdzVqev3V4Jpw3P"),
            Pubkey.from_string("5d2QJ6u2NveZufmJ4noHja5EHs3Bv1DUMPLG5xfasSVs"),
            Pubkey.from_string("11111111111111111111111111111111"),
            Pubkey.from_string("5SSkXsEKQepHHAewytPVwdej4epN1nxgLVM84L4KXgy7"),
            Pubkey.from_string("11111111111111111111111111111111"),
            Pubkey.from_string("JB43m9cTZVb4LcEfBA1Bf49nz4cod2Xt5i2HRAAMQvdo"),
            Pubkey.from_string("6YcgjoTUTeafyFC5C6vgFCVoM69qgpj966PhCojwiS4Z"),
            Pubkey.from_string("qjoUa8fC1DjsnwVwUCJQcVDJVYUhAnrqz3B9FSxVPAm"),
            Pubkey.from_string("5kSQn4o4S4mRR48sLZ5kijvCjnnD4NbwPqaBaQYvC6wk"),
            Pubkey.from_string("7Nm1pAysuk38D6uiU83wXfegzruqohVNwUippwSZNx71"),
            Pubkey.from_string("V6Y9bCVTTwjBDYAJ6ixMWWFK7RKZcQeUcpufZaM3DDB"),
            Pubkey.from_string("5yiNVYs4xpC26yY9K3DYTRvdFkdjBSnkyeSAA3Huf4Q2"),
            Pubkey.from_string("CjwQEszLimhf3FkwqcB5rGppZS4ADLeCmtPvcRefv5qi"),
            Pubkey.from_string("HM95yakZehLhBkVx2a5mwzTTf4UeqfDeSno17NstQKzd"),
            Pubkey.from_string("GHSuFYv8SicmVC3c5DeKgFhjDdegNH552srPFufFysDU"),
            Pubkey.from_string("HT9nmiVtAbLp3GFtfrhGCPGz1tBcvHyFz43pf8rtDArs"),
            Pubkey.from_string("J8yPaachUYRo1qbtdvkQU6Ee2vzEt3JiYRncwkfoB3iZ"),
            Pubkey.from_string("D9mSSsfXp9SUrSCrmPA8sDJ42GMn65Z5tZr9wtwbHEBy"),
            Pubkey.from_string("AYUthyJDg6mvBWjGfEGYcnjRbw3TZbYCnEQyqUiPyw57"),
            Pubkey.from_string("D7g6CGNVzebzzH1t98tBJKjA6FokiuXKaX3QVF53ZqcC"),
            Pubkey.from_string("DLzvYrHo4bT8PoSwhxDWKhH9ZNsnPbE5cgCx1coCpeR6"),
            Pubkey.from_string("2CtvLehungAsAcYhqnhLEoxuTaM5WqHrycn2YNpmnC4z"),
            Pubkey.from_string("9bGCvH7MwfwyKuNzgfGRXU8sW9jWtFMJf4FmUsQxk13Z"),
            Pubkey.from_string("8pQvXjUf9qq6Lsk9WE8tqtPXxgf4Yav32DneQDLaJr9k"),
            Pubkey.from_string("Gxc3KYEgasnSbevj9GnzwDk2i82BRHACQ8HWtgA1VXUk"),
            Pubkey.from_string("2fgyaTU1BeSk3LWbEh9p3D1H29hEt1wyy8qchKNGxBhg"),
            Pubkey.from_string("FDXX8a9asz32d6EteL6Wnra71YjT45TuGHy8Y8dPntbq"),
            Pubkey.from_string("JCk2bU5xNDUBk55W5BoEgNeKqssXdj15vLqhuzQEEiCU"),
            Pubkey.from_string("HYv5Vb7G3849dYN8RG2pQx7awXiEk3nE9F3AGWnedffs"),
            Pubkey.from_string("5RDbC6wYGhc1AjCCArwggtbAyPUse6yMcq3fgSzZqY5E"),
            Pubkey.from_string("2CCypxRdB6tCackur4De72ZSnsNcicfSjWYZwAcVzEZ9"),
            Pubkey.from_string("J5e19jEz625k7K5K3yinpMoHNzznv8smCemvKAYCWLMH"),
            Pubkey.from_string("7QDHQhNjMzuKKk9T5vsdZ5hTfLu2P1pUGk2TKJKFd8hn"),
            Pubkey.from_string("DYXqC6UGPwsz5W3jaFAJtpbBsvJpBSi34W6cMn7ZiL6m"),
            Pubkey.from_string("2YBnvgCAh6WcXrs5Wz2uhi2nMuEVDXE28mQ5AmHjRaJQ"),
            Pubkey.from_string("wkTZtYAo75bpqqy2J5A6emsLK69EYG8otvJipAbiJF9"),
            Pubkey.from_string("FXgqEhVGX82dsyKMAgR2JCRaC9FNH4ih611bJftmD3tG"),
            Pubkey.from_string("B3iV7Y5S9Xqu6cTECpBCAHtL2nwvZntD2d7F1zbpbQc3"),
            Pubkey.from_string("J7E7gv1iofWZg8b37RVnKy2i1FAnCRBvgk54YpVtZ4WS"),
            Pubkey.from_string("EPf7hymYW7bnwiBYGTRF4Ji2jJ2yTdn4XMTpR6N4zsGA"),
            Pubkey.from_string("DNXXx5yaQm6Kd4embzqHW94kjjypxzZ8zCQthMvmHDq1"),
            Pubkey.from_string("6VUvm4i9T2w9CXzLKaUDNTZ4KmK3xKCWz95xEoyKjjE8"),
            Pubkey.from_string("Dd3K16uKGDa2Xc83NYQE9jpcMNZ9GRqmqggkkpXo4x4y"),
            Pubkey.from_string("FUsWmTk7PaEvXCHFZH6XMbimgqbVkiKPXd35tHWo5vm4"),
            Pubkey.from_string("Fc7FU4VATyYuvkhp6MuPUHWYHC9EBFknN5TTHs3wGjyn"),
            Pubkey.from_string("CW5zwzmhQn9Pqjzxjdzr4AdmVSoehBQ2RdfyppU37uDE"),
            Pubkey.from_string("Cvqd7RtB4bf9SJpmWwNd6GVXvtAX4m4DtSR5u2XahTQf"),
            Pubkey.from_string("67XFnD7YuBfd52qPrD5FuasrQW21jLd4jxnDfozyeSvx"),
            Pubkey.from_string("3H7nvQAJBzsXVPEyzpuL4kB3tPz5rs7XVxuxVLV7YDNy"),
            Pubkey.from_string("JEDqstU8skdUF6bGsMzZaa17Tfuf3asxQPuLVXqmWqyt"),
            Pubkey.from_string("2Ux9NzgadfR2PHyrSauzWLaNr99syCRKriwF4mnEqPUv"),
            Pubkey.from_string("4NEK9db8v9PZC5RnzSEh596jNwfRMH6QcQQG2ZhqkStb"),
            Pubkey.from_string("2mRZPMv67hz1QocPQ7JS5sEp8uD9u9TfdLoZzf8LciiZ"),
            Pubkey.from_string("G3ytGgn7MzH5DpLCSBshkV4xQe1G4JBbTu7c9CCWA9U"),
            Pubkey.from_string("J5t2VsFLSXGTaA1o2qCujduJ9dnHLok9QhYUroN2fc6X"),
            Pubkey.from_string("6UEwXk5ixtC6rDJG7X2X6MLNwiAyjGZZyN6L1JG63n4H"),
            Pubkey.from_string("FkHx7byyFDGheXCcmVxdavTDteotaDJZk8DBNBm2Maeb"),
            Pubkey.from_string("zXEzviQpq8u89oAKdU7r3vHHxXpxC2547AoLzxCKSJ3"),
            Pubkey.from_string("93DGPpMR6w3gqbrbpZrChqw8vihgHibmdejB4Q1p4DbP"),
            Pubkey.from_string("Ag4cqXXAAxa9NCZcFgQuoeDkvtSbxwe6qaEfjw4aRnUH"),
        ],
    ),
    Network.MAINNET: AddressLookupTableAccount(
        key=Pubkey.from_string("7Xjap9vrWush2kY3HV691U2R3b4CZgyafWcCDJiZ1zVR"),
        addresses=[
            Pubkey.from_string("8eExPiLp47xbSDYkbuem4qnLUpbLTfZBeFuEJoh6EUr2"),
            Pubkey.from_string("BbKFezrmKD83PeVh74958MzgFAue1pZptipSNLz5ccpk"),
            Pubkey.from_string("zDEXqXEG7gAyxb1Kg9mK5fPnUdENCGKzWrM21RMdWRq"),
            Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"),
            Pubkey.from_string("AVNMK6wiGfppdQNg9WKfMRBXefDPGZFh2f3o1fRbgN8n"),
            Pubkey.from_string("SysvarRent111111111111111111111111111111111"),
            Pubkey.from_string("HEvSKofvBgfaexv23kMabbYqxasxU3mQ4ibBMEmJWHny"),
            Pubkey.from_string("H6ARHf6YXhGYeQfUzQNGk6rDNnLBQKrenN712K4AQJEG"),
            Pubkey.from_string("CH31Xns5z3M1cTAbKW34jcxPPciazARpijcHj9rxtemt"),
            Pubkey.from_string("GVXRSBjFk6e6J3NbVPXohDJetcTjaeeuykUpbQF8UoMU"),
            Pubkey.from_string("Cv4T27XbjVoKUYwP72NQQanvZeA7W4YF9L4EnYT9kx5o"),
            Pubkey.from_string("JBu1AL4obBcCMqKBBxhpWCNUt136ijcuMZLFvTP7iWdB"),
            Pubkey.from_string("716hFAECqotxcXcj8Hs8nr7AG6q9dBw2oX3k3M8V7uGq"),
            Pubkey.from_string("FNNvb1AFDnDVPkocEri8mWbJ1952HQZtFLuwPiUjSJQ"),
            Pubkey.from_string("11111111111111111111111111111111"),
            Pubkey.from_string("5HRrdmghsnU3i2u5StaKaydS7eq3vnKVKwXMzCNKsc4C"),
            Pubkey.from_string("11111111111111111111111111111111"),
            Pubkey.from_string("JE6d41JRokZAMUEAznV8JP4h7i6Ain6CyJrQuweRipFU"),
            Pubkey.from_string("EaNR74nCjrYyNDsuoWmq19pH76QSd1nuTzvJSr3RDQ6x"),
            Pubkey.from_string("3rjBffJkFa9zdGr9xmTVxF3y6nL6iL1pASVzym7s5FGr"),
            Pubkey.from_string("HcjrQKnbqKJuxHDLNM9LJPyxcQs237waNZXW7RwgvAps"),
            Pubkey.from_string("Ec4xsLLgLc4wM5c19ZSvazE7M9Rtk2T6RzddNcSQKYGu"),
            Pubkey.from_string("BEjGhNFnKT5weGtpBoFs5Y1mDN47Ntvag5aMV59nZRpk"),
            Pubkey.from_string("CHBUBfU3zscTNsihdK3x44TbpTza1hwcsTUaZfk751b5"),
            Pubkey.from_string("GYm6qTFwkGJx2ywetEuYrjHjhzVFCM2TwayyqS1HUPLG"),
            Pubkey.from_string("7aXkF7AZE2D3h128eNJ7VVp72HCV1izjKFsJ8uNWtCFN"),
            Pubkey.from_string("BKt2FdgBahn77joeawhNidswFxfgasPYCHWghRL4AKBR"),
            Pubkey.from_string("2ZLdhFsrkAtdn9Kud4SZvqchQFvn5jVCHUdJ83vumKyR"),
            Pubkey.from_string("J5DTCqRAjX1FyzoP2A2HVmmaXuG971HJ8X1Z8Rvvd8uT"),
            Pubkey.from_string("6JSdqUr24mBt4MCQrZHRoSfeZbjgALx4MQunZwD8Uarg"),
            Pubkey.from_string("4K1zxqAZn7bGAPN26W5mUaHrLMSpCk45gT4qVXwmfh39"),
            Pubkey.from_string("8JSbFw4YT3bzpbbHs1wKmRCSAmKucAba7XSUWj1p8xK5"),
            Pubkey.from_string("483SmqGQVxw3WDwcewMYHqC3Mu7ENxfTQJQtTR3ttpi7"),
            Pubkey.from_string("DbzL5mT4nBaxuAs8ti4UeT2qougRBdujxa7GhLndM5Jz"),
            Pubkey.from_string("7M9xhY2ARnrkCaBK5SNM3Lyd3FdbTu2EWBwG4TQcqpsv"),
            Pubkey.from_string("5mvaZWcZa4gB3JZptZuFAJnmDFfo1JovhqTkkfEcsryD"),
            Pubkey.from_string("DboTtkWW3KPvT14fag8N6iDUyDXaT8FeBszGV9xdfBx2"),
            Pubkey.from_string("DhMH8oRQoAAb6poHVsvCqq3NCMj6aKUH2tGQG5Lo4bCg"),
            Pubkey.from_string("63DZkAzoDXmzGzn9esoWSYpMLo4YB9oPHXreHKwuu4HA"),
            Pubkey.from_string("J8x6y5G7GmTkuKTbbCAnfhn72vaUU2qsB6je9oKFigHM"),
            Pubkey.from_string("A7D8zuxAmtui3XKz2VcxthAZ5HuwLbN74rrDapMJ3Z5d"),
            Pubkey.from_string("CzK26LWpoU9UjSrZkVu97oZj63abJrNv1zp9Hy2zZdy5"),
            Pubkey.from_string("CaqN8gomPEB1My2czRhetrj5ESKP3VRR3kwKhXtGxohG"),
            Pubkey.from_string("4oAjuVLt5N9au2X3bhYLSi9LRqtk4caBvSPQRwhXuEKP"),
            Pubkey.from_string("9YVE9r9cHFZNwm91p3Js8NBWVznesLSM8FZyswG2MG1B"),
            Pubkey.from_string("DecjLCYjb7jdDp2UqA2MS4xjgDjZfvdgMjvkRW7oWs9L"),
            Pubkey.from_string("8SH6uJe5rV13APZFQvkGdXPwTyeiLv67XTnv3EeYff3B"),
            Pubkey.from_string("DhWWXYK2DSnCdK5rkAxJBkGb1SBR49RHpfHj2u1vobCJ"),
            Pubkey.from_string("5Ehp2LtTRmjug39GphXhFEeguz7hGeg41N1U49wU8Kov"),
            Pubkey.from_string("2Stzi7XE3btUQXaauTVB9brPAtPmGnrEDSJmp3w5VY2j"),
            Pubkey.from_string("J6feTwcYDydW71Dp9qgfW7Mu9qk3qDRrDZAWV8NMVh9x"),
            Pubkey.from_string("EPf7hymYW7bnwiBYGTRF4Ji2jJ2yTdn4XMTpR6N4zsGA"),
            Pubkey.from_string("DNXXx5yaQm6Kd4embzqHW94kjjypxzZ8zCQthMvmHDq1"),
            Pubkey.from_string("6VUvm4i9T2w9CXzLKaUDNTZ4KmK3xKCWz95xEoyKjjE8"),
            Pubkey.from_string("Dd3K16uKGDa2Xc83NYQE9jpcMNZ9GRqmqggkkpXo4x4y"),
            Pubkey.from_string("GqCVQuGMf8YkiaJkSrD98D3WZxFfktcKzAwdBEQsx1th"),
            Pubkey.from_string("2BpEtArGNotp97DjVKwhYZ86WEq2Y5bLtGGDvicuJ9br"),
            Pubkey.from_string("6nkbQueQanrsg2JVvqoZ6zPuZsiL3oLmTk7xgEpRAirD"),
            Pubkey.from_string("7T1qeETZ6ZpbKJ7wmrWoxhewLaJRW9H4EgMnR1tgHYRQ"),
            Pubkey.from_string("6S6WYL1mQFmVxsf3ft5MEH8hzxJA1LcUDzgwdJDj3yxc"),
            Pubkey.from_string("GuNWJSV4k95FZdwhAcjdaPGGoh9cArc27yV4P54QwWdg"),
            Pubkey.from_string("J18LXTGe2cPgpLKSCwXiG6tYkjcqEiaMUznM19Q8faVL"),
            Pubkey.from_string("gP91avgCrV9KB2ATgRtMCNN2AN7oU9hK1frENe17QkR"),
            Pubkey.from_string("DUtwwtnNBzpPTnzUiz23vqH93gd1qwEzGisyC5bLwCBv"),
            Pubkey.from_string("A4SayHehmafd6DNjrzY3L9ddQEGw4UHV3LMcKfmyPcMT"),
            Pubkey.from_string("8if1NcDaif8dxM3Ct6rRTQEj9GFwukadU1MDQHqCnw9h"),
            Pubkey.from_string("GnSRgncxFbtxqZ4HmfnF6daCmgkc8tuQz9i37hUmwV5t"),
            Pubkey.from_string("6JjDgGzqzU6Az7ZmTARAvBSwBxfXsqbVG3Rc9JGU9i4L"),
            Pubkey.from_string("GQXdvh4dHvENTFi2CfVrh77aQD1Y6V7HMNYNCuvgbSuc"),
            Pubkey.from_string("EC2PCjwcuBBFHp6gpAKa3kdpLVrQxdy7kgr7p4wPy8Vw"),
            Pubkey.from_string("7j1N5UiXLFxaxFWq5tzZc5R3sjPHcF7jqfHJgAtE74q8"),
            Pubkey.from_string("Eqt3anUy8nqDvzJaNvWvqBM32Ln4UUnLkfvdd9Ztfj81"),
        ],
    ),
}



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class BidJSON(typing.TypedDict):
    kind: typing.Literal["Bid"]


class AskJSON(typing.TypedDict):
    kind: typing.Literal["Ask"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Bid:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Bid"

    @classmethod
    def to_json(cls) -> BidJSON:
        return BidJSON(
            kind="Bid",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Bid": {},
        }


@dataclass
class Ask:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Ask"

    @classmethod
    def to_json(cls) -> AskJSON:
        return AskJSON(
            kind="Ask",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Ask": {},
        }


SideKind = typing.Union[Uninitialized, Bid, Ask]
SideJSON = typing.Union[UninitializedJSON, BidJSON, AskJSON]


def from_decoded(obj: dict) -> SideKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Bid" in obj:
        return Bid()
    if "Ask" in obj:
        return Ask()
    raise ValueError("Invalid enum object")


def from_json(obj: SideJSON) -> SideKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Bid":
        return Bid()
    if obj["kind"] == "Ask":
        return Ask()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Uninitialized" / borsh.CStruct(), "Bid" / borsh.CStruct(), "Ask" / borsh.CStruct())



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdateInterestRateArgsJSON(typing.TypedDict):
    expiry_index: int
    interest_rate: int


@dataclass
class UpdateInterestRateArgs:
    layout: typing.ClassVar = borsh.CStruct("expiry_index" / borsh.U8, "interest_rate" / borsh.I64)
    expiry_index: int
    interest_rate: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateInterestRateArgs":
        return cls(expiry_index=obj.expiry_index, interest_rate=obj.interest_rate)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"expiry_index": self.expiry_index, "interest_rate": self.interest_rate}

    def to_json(self) -> UpdateInterestRateArgsJSON:
        return {"expiry_index": self.expiry_index, "interest_rate": self.interest_rate}

    @classmethod
    def from_json(cls, obj: UpdateInterestRateArgsJSON) -> "UpdateInterestRateArgs":
        return cls(expiry_index=obj["expiry_index"], interest_rate=obj["interest_rate"])



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from construct import Container
from solders.pubkey import Pubkey

from . import asset


class UpdateZetaPricingPubkeysArgsJSON(typing.TypedDict):
    asset: asset.AssetJSON
    oracle: str
    oracle_backup_feed: str
    market: str
    perp_sync_queue: str
    zeta_group_key: str


@dataclass
class UpdateZetaPricingPubkeysArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "asset" / asset.layout,
        "oracle" / BorshPubkey,
        "oracle_backup_feed" / BorshPubkey,
        "market" / BorshPubkey,
        "perp_sync_queue" / BorshPubkey,
        "zeta_group_key" / BorshPubkey,
    )
    asset: asset.AssetKind
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    market: Pubkey
    perp_sync_queue: Pubkey
    zeta_group_key: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateZetaPricingPubkeysArgs":
        return cls(
            asset=asset.from_decoded(obj.asset),
            oracle=obj.oracle,
            oracle_backup_feed=obj.oracle_backup_feed,
            market=obj.market,
            perp_sync_queue=obj.perp_sync_queue,
            zeta_group_key=obj.zeta_group_key,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "asset": self.asset.to_encodable(),
            "oracle": self.oracle,
            "oracle_backup_feed": self.oracle_backup_feed,
            "market": self.market,
            "perp_sync_queue": self.perp_sync_queue,
            "zeta_group_key": self.zeta_group_key,
        }

    def to_json(self) -> UpdateZetaPricingPubkeysArgsJSON:
        return {
            "asset": self.asset.to_json(),
            "oracle": str(self.oracle),
            "oracle_backup_feed": str(self.oracle_backup_feed),
            "market": str(self.market),
            "perp_sync_queue": str(self.perp_sync_queue),
            "zeta_group_key": str(self.zeta_group_key),
        }

    @classmethod
    def from_json(cls, obj: UpdateZetaPricingPubkeysArgsJSON) -> "UpdateZetaPricingPubkeysArgs":
        return cls(
            asset=asset.from_json(obj["asset"]),
            oracle=Pubkey.from_string(obj["oracle"]),
            oracle_backup_feed=Pubkey.from_string(obj["oracle_backup_feed"]),
            market=Pubkey.from_string(obj["market"]),
            perp_sync_queue=Pubkey.from_string(obj["perp_sync_queue"]),
            zeta_group_key=Pubkey.from_string(obj["zeta_group_key"]),
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class PositionMovementArgJSON(typing.TypedDict):
    index: int
    size: int


@dataclass
class PositionMovementArg:
    layout: typing.ClassVar = borsh.CStruct("index" / borsh.U8, "size" / borsh.I64)
    index: int
    size: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "PositionMovementArg":
        return cls(index=obj.index, size=obj.size)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"index": self.index, "size": self.size}

    def to_json(self) -> PositionMovementArgJSON:
        return {"index": self.index, "size": self.size}

    @classmethod
    def from_json(cls, obj: PositionMovementArgJSON) -> "PositionMovementArg":
        return cls(index=obj["index"], size=obj["size"])



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class LimitJSON(typing.TypedDict):
    kind: typing.Literal["Limit"]


class PostOnlyJSON(typing.TypedDict):
    kind: typing.Literal["PostOnly"]


class FillOrKillJSON(typing.TypedDict):
    kind: typing.Literal["FillOrKill"]


class ImmediateOrCancelJSON(typing.TypedDict):
    kind: typing.Literal["ImmediateOrCancel"]


class PostOnlySlideJSON(typing.TypedDict):
    kind: typing.Literal["PostOnlySlide"]


@dataclass
class Limit:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Limit"

    @classmethod
    def to_json(cls) -> LimitJSON:
        return LimitJSON(
            kind="Limit",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Limit": {},
        }


@dataclass
class PostOnly:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "PostOnly"

    @classmethod
    def to_json(cls) -> PostOnlyJSON:
        return PostOnlyJSON(
            kind="PostOnly",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PostOnly": {},
        }


@dataclass
class FillOrKill:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "FillOrKill"

    @classmethod
    def to_json(cls) -> FillOrKillJSON:
        return FillOrKillJSON(
            kind="FillOrKill",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FillOrKill": {},
        }


@dataclass
class ImmediateOrCancel:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "ImmediateOrCancel"

    @classmethod
    def to_json(cls) -> ImmediateOrCancelJSON:
        return ImmediateOrCancelJSON(
            kind="ImmediateOrCancel",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ImmediateOrCancel": {},
        }


@dataclass
class PostOnlySlide:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "PostOnlySlide"

    @classmethod
    def to_json(cls) -> PostOnlySlideJSON:
        return PostOnlySlideJSON(
            kind="PostOnlySlide",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PostOnlySlide": {},
        }


OrderTypeKind = typing.Union[Limit, PostOnly, FillOrKill, ImmediateOrCancel, PostOnlySlide]
OrderTypeJSON = typing.Union[LimitJSON, PostOnlyJSON, FillOrKillJSON, ImmediateOrCancelJSON, PostOnlySlideJSON]


def from_decoded(obj: dict) -> OrderTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Limit" in obj:
        return Limit()
    if "PostOnly" in obj:
        return PostOnly()
    if "FillOrKill" in obj:
        return FillOrKill()
    if "ImmediateOrCancel" in obj:
        return ImmediateOrCancel()
    if "PostOnlySlide" in obj:
        return PostOnlySlide()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderTypeJSON) -> OrderTypeKind:
    if obj["kind"] == "Limit":
        return Limit()
    if obj["kind"] == "PostOnly":
        return PostOnly()
    if obj["kind"] == "FillOrKill":
        return FillOrKill()
    if obj["kind"] == "ImmediateOrCancel":
        return ImmediateOrCancel()
    if obj["kind"] == "PostOnlySlide":
        return PostOnlySlide()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Limit" / borsh.CStruct(),
    "PostOnly" / borsh.CStruct(),
    "FillOrKill" / borsh.CStruct(),
    "ImmediateOrCancel" / borsh.CStruct(),
    "PostOnlySlide" / borsh.CStruct(),
)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.sysvar import CLOCK


@dataclass
class Clock:
    layout: typing.ClassVar = borsh.CStruct(
        "slot" / borsh.U64,
        "epoch_start_timestamp" / borsh.I64,
        "epoch" / borsh.U64,
        "leader_schedule_epoch" / borsh.U64,
        "unix_timestamp" / borsh.I64,
    )
    slot: int
    epoch_start_timestamp: int
    epoch: int
    leader_schedule_epoch: int
    unix_timestamp: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        commitment: typing.Optional[Commitment] = None,
    ) -> typing.Optional["Clock"]:
        resp = await conn.get_account_info(CLOCK, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    def decode(cls, data: bytes) -> "Clock":
        dec = Clock.layout.parse(data)
        return cls(
            slot=dec.slot,
            epoch_start_timestamp=dec.epoch_start_timestamp,
            epoch=dec.epoch,
            leader_schedule_epoch=dec.leader_schedule_epoch,
            unix_timestamp=dec.unix_timestamp,
        )



import asyncio
import json
import logging
import time
import traceback
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Union, cast

import based58
import websockets
from anchorpy import Event, Provider, Wallet
from anchorpy.provider import DEFAULT_OPTIONS
from construct import Container
from deprecated import deprecated
from jsonrpcclient import request
from solana.blockhash import BlockhashCache
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment, Confirmed
from solana.rpc.core import RPCException
from solana.rpc.types import TxOpts
from solana.rpc.websocket_api import SolanaWsClientProtocol, connect
from solders.instruction import Instruction
from solders.message import MessageV0
from solders.pubkey import Pubkey
from solders.rpc.config import RpcTransactionLogsFilterMentions
from solders.transaction import VersionedTransaction

from zetamarkets_py import constants, pda, utils
from zetamarkets_py.events import (
    CancelOrderEvent,
    LiquidationEvent,
    OrderCompleteEvent,
    PlaceOrderEvent,
    PlaceOrderEventWithArgs,
    TradeEvent,
    ZetaEnrichedEvent,
    ZetaEvent,
)
from zetamarkets_py.exchange import Exchange
from zetamarkets_py.orderbook import Orderbook
from zetamarkets_py.serum_client.accounts.orderbook import OrderbookAccount
from zetamarkets_py.solana_client.accounts.clock import CLOCK, Clock
from zetamarkets_py.types import (
    Asset,
    Network,
    OrderArgs,
    OrderCompleteType,
    OrderOptions,
    Position,
    Side,
)
from zetamarkets_py.zeta_client.accounts.cross_margin_account import CrossMarginAccount
from zetamarkets_py.zeta_client.errors import from_tx_error
from zetamarkets_py.zeta_client.instructions import (
    cancel_all_market_orders,
    cancel_order,
    deposit_v2,
    initialize_cross_margin_account,
    initialize_cross_margin_account_manager,
    initialize_open_orders_v3,
    place_perp_order_v3,
)

# TODO: add docstrings for most methods
# TODO: implement better error handling and tracebacks
# TODO: implement withdraw and liquidation
# TODO: implement priority fees to exchange
# TODO: implement clock subscription


@dataclass
class Client:
    """
    Cross margin client
    """

    provider: Provider
    network: Network
    connection: AsyncClient
    endpoint: str
    ws_endpoint: str
    exchange: Exchange

    margin_account: Optional[CrossMarginAccount]

    _margin_account_address: Optional[Pubkey]
    _open_orders_addresses: Optional[dict[Asset, Pubkey]]
    _margin_account_manager_address: Optional[Pubkey]
    _user_usdc_address: Optional[Pubkey]
    _combined_vault_address: Pubkey
    _combined_socialized_loss_address: Pubkey
    _logger: logging.Logger
    _account_exists_cache: dict[Pubkey, bool] = field(default_factory=dict)

    @classmethod
    async def load(
        cls,
        endpoint: Optional[str] = None,
        ws_endpoint: Optional[str] = None,
        commitment: Commitment = Confirmed,
        wallet: Optional[Wallet] = None,
        assets: list[Asset] = Asset.all(),
        tx_opts: TxOpts = DEFAULT_OPTIONS,
        network: Network = Network.MAINNET,
        log_level: int = logging.CRITICAL,
        blockhash_cache: Union[BlockhashCache, bool] = False,
    ):
        """
        Create a new client
        """
        logger = utils.create_logger(f"{__name__}.{cls.__name__}", log_level)

        if endpoint is None:
            endpoint = utils.cluster_endpoint(network)
        if ws_endpoint is None:
            ws_endpoint = utils.http_to_ws(endpoint)
        connection = AsyncClient(endpoint=endpoint, commitment=commitment, blockhash_cache=blockhash_cache)
        exchange = await Exchange.load(
            network=network,
            connection=connection,
            assets=assets,
        )
        if wallet is None:
            wallet = Wallet.dummy()
            logger.warning("Client in read-only mode, pass in `wallet` to enable transactions")
            _margin_account_manager_address = None
            _user_usdc_address = None
            _margin_account_address = None
            margin_account = None
            _open_orders_addresses = None
        else:
            _margin_account_manager_address = pda.get_cross_margin_account_manager_address(
                exchange.program_id, wallet.public_key
            )
            _user_usdc_address = pda.get_associated_token_address(wallet.public_key, constants.USDC_MINT[network])
            _margin_account_address = pda.get_margin_account_address(exchange.program_id, wallet.public_key, 0)
            margin_account = await CrossMarginAccount.fetch(
                connection, _margin_account_address, connection.commitment, exchange.program_id
            )
            _open_orders_addresses = {}
            for asset in assets:
                open_orders_address = pda.get_open_orders_address(
                    exchange.program_id,
                    constants.MATCHING_ENGINE_PID[network],
                    exchange.markets[asset].address,
                    _margin_account_address,
                )
                _open_orders_addresses[asset] = open_orders_address
        provider = Provider(
            connection,
            wallet,
            tx_opts,
        )

        # additional addresses to cache
        _combined_vault_address = pda.get_combined_vault_address(exchange.program_id)
        _combined_socialized_loss_address = pda.get_combined_socialized_loss_address(exchange.program_id)

        return cls(
            provider,
            network,
            connection,
            endpoint,
            ws_endpoint,
            exchange,
            margin_account,
            _margin_account_address,
            _open_orders_addresses,
            _margin_account_manager_address,
            _user_usdc_address,
            _combined_vault_address,
            _combined_socialized_loss_address,
            logger,
        )

    async def _check_user_usdc_account_exists(self):
        if self._margin_account_manager_address in self._account_exists_cache:
            return self._account_exists_cache[self._margin_account_manager_address]
        resp = await self.connection.get_account_info(self._user_usdc_address)
        exists = resp.value is not None
        self._account_exists_cache[self._margin_account_manager_address] = exists
        return exists

    async def _check_margin_account_manager_exists(self):
        if self._margin_account_manager_address in self._account_exists_cache:
            return self._account_exists_cache[self._margin_account_manager_address]
        resp = await self.connection.get_account_info(self._user_usdc_address)
        exists = resp.value is not None
        self._account_exists_cache[self._margin_account_manager_address] = exists
        return exists

    async def _check_margin_account_exists(self):
        if self._margin_account_manager_address in self._account_exists_cache:
            return self._account_exists_cache[self._margin_account_manager_address]
        account = await CrossMarginAccount.fetch(self.connection, self._margin_account_address)
        exists = account is not None
        self._account_exists_cache[self._margin_account_manager_address] = exists
        self.margin_account = account
        return exists

    async def _check_open_orders_account_exists(self, asset: Asset):
        if self._open_orders_addresses is None:
            raise Exception("Open orders accounts not loaded, cannot check if account exists")
        open_orders_address = self._open_orders_addresses[asset]
        if open_orders_address in self._account_exists_cache:
            return self._account_exists_cache[open_orders_address]
        resp = await self.connection.get_account_info(open_orders_address)
        exists = resp.value is not None
        self._account_exists_cache[open_orders_address] = exists
        return exists

    @deprecated(version="0.1.21", reason="You should use fetch_margin_state instead")
    async def fetch_balance(self):
        margin_account = await self.margin_account.fetch(
            self.connection,
            self._margin_account_address,
            program_id=self.exchange.program_id,
        )
        balance = utils.convert_fixed_int_to_decimal(margin_account.balance)
        return balance

    @deprecated(version="0.1.21", reason="You should use fetch_margin_state instead")
    async def fetch_position(self, asset: Asset):
        if self.margin_account is None or self._margin_account_address is None:
            raise Exception("Margin account not loaded, cannot fetch position")
        margin_account = await self.margin_account.fetch(
            self.connection, self._margin_account_address, program_id=self.exchange.program_id
        )
        if margin_account is None:
            raise Exception("Margin account not found, cannot fetch position")
        position = Position(
            utils.convert_fixed_lot_to_decimal(margin_account.product_ledgers[asset.to_index()].position.size),
            utils.convert_fixed_int_to_decimal(
                margin_account.product_ledgers[asset.to_index()].position.cost_of_trades
            ),
        )
        return position

    async def fetch_margin_state(self):
        if self.margin_account is None or self._margin_account_address is None:
            raise Exception("Margin account not loaded, cannot fetch margin account state")
        margin_account = await self.margin_account.fetch(
            self.connection, self._margin_account_address, program_id=self.exchange.program_id
        )
        if margin_account is None:
            raise Exception("Margin account not found, cannot fetch margin state")
        balance = utils.convert_fixed_int_to_decimal(margin_account.balance)
        positions = {
            a: Position(
                utils.convert_fixed_lot_to_decimal(margin_account.product_ledgers[a.to_index()].position.size),
                utils.convert_fixed_int_to_decimal(
                    margin_account.product_ledgers[a.to_index()].position.cost_of_trades
                ),
            )
            for a in self.exchange.assets
        }
        return balance, positions

    async def fetch_open_orders(self, asset: Asset):
        if self._open_orders_addresses is None:
            raise Exception("Open orders accounts not loaded, cannot fetch open orders")
        oo = await self.exchange.markets[asset].load_orders_for_owner(self._open_orders_addresses[asset])
        return oo

    async def fetch_clock(self):
        clock = await Clock.fetch(self.connection)
        if clock is None:
            raise Exception("Clock not found, cannot fetch clock")
        return clock

    async def _account_subscribe(
        self,
        address: Pubkey,
        commitment: Commitment,
        max_retries: int = 3,
        encoding: str = "base64+zstd",
    ) -> AsyncIterator[Tuple[bytes, int]]:
        retries = max_retries
        while True:
            try:
                async with connect(self.ws_endpoint) as ws:
                    solana_ws: SolanaWsClientProtocol = cast(SolanaWsClientProtocol, ws)
                    await solana_ws.account_subscribe(
                        address,
                        commitment=commitment,
                        encoding=encoding,
                    )
                    first_resp = await solana_ws.recv()
                    subscription_id = cast(int, first_resp[0].result)
                    async for msg in ws:
                        try:
                            slot = int(msg[0].result.context.slot)
                            account_bytes = cast(bytes, msg[0].result.value.data)  # type: ignore
                            yield account_bytes, slot
                        except Exception:
                            self._logger.error(f"Error processing account data: {traceback.format_exc()}")
                            break
                    await solana_ws.account_unsubscribe(subscription_id)
            except asyncio.CancelledError:
                self._logger.info("WebSocket subscription task cancelled.")
                break
            except websockets.exceptions.ConnectionClosed:
                self._logger.error("WebSocket connection closed unexpectedly. Attempting to reconnect...")
            except Exception as e:
                self._logger.error(f"Error subscribing to {self.__class__.__name__}: {e}. Retrying...")
                retries -= 1
                if retries <= 0:
                    self._logger.error("Max retries reached. Unable to subscribe to account.")
                    break
                await asyncio.sleep(2)  # Pause for a while before retrying

    async def subscribe_orderbook(
        self, asset: Asset, side: Side, commitment: Optional[Commitment] = None, max_retries: int = 3
    ) -> AsyncIterator[Tuple[Orderbook, int]]:
        commitment = commitment or self.connection.commitment
        address = (
            self.exchange.markets[asset]._market_state.bids
            if side == Side.Bid
            else self.exchange.markets[asset]._market_state.asks
        )

        self._logger.info(f"Subscribing to Orderbook:{side}.")
        async for account_bytes, slot in self._account_subscribe(address, commitment, max_retries):
            account = OrderbookAccount.decode(account_bytes)
            orderbook = Orderbook(side, account, self.exchange.markets[asset]._market_state)
            yield orderbook, slot

    async def subscribe_clock(
        self, commitment: Optional[Commitment] = None, max_retries: int = 3
    ) -> AsyncIterator[Tuple[Clock, int]]:
        commitment = commitment or self.connection.commitment

        self._logger.info("Subscribing to Clock")
        async for account_bytes, slot in self._account_subscribe(CLOCK, commitment, max_retries):
            clock = Clock.decode(account_bytes)
            yield clock, slot

    # TODO: maybe at some point support subscribing to all exchange events, not just margin account
    async def subscribe_events(
        self,
        commitment: Optional[Commitment] = None,
        max_retries: int = 3,
    ) -> AsyncIterator[Tuple[List[ZetaEvent], int]]:
        if self._margin_account_address is None:
            raise Exception("Margin account not loaded, cannot subscribe to events")
        commitment = commitment or self.connection.commitment
        retries = max_retries
        while retries > 0:
            try:
                async with connect(self.ws_endpoint) as ws:
                    solana_ws: SolanaWsClientProtocol = cast(SolanaWsClientProtocol, ws)
                    # Subscribe to logs that mention the margin account
                    await solana_ws.logs_subscribe(
                        commitment=commitment,
                        filter_=RpcTransactionLogsFilterMentions(self._margin_account_address),
                    )
                    first_resp = await solana_ws.recv()
                    subscription_id = cast(int, first_resp[0].result)
                    async for msg in ws:
                        try:
                            events, slot = self.parse_event_payload(msg)
                            if len(events) > 0:
                                yield events, slot

                        except Exception:
                            self._logger.error(f"Error processing event data: {traceback.format_exc()}")
                            break
                    await solana_ws.logs_unsubscribe(subscription_id)
            except asyncio.CancelledError:
                self._logger.info("WebSocket subscription task cancelled.")
                break
            except websockets.exceptions.ConnectionClosed:
                self._logger.error("WebSocket connection closed unexpectedly. Attempting to reconnect...")
            except Exception as e:
                self._logger.error(f"Error subscribing to {self.__class__.__name__}: {e}. Retrying...")
                retries -= 1
                if retries <= 0:
                    self._logger.error("Max retries reached. Unable to subscribe to events.")
                    break
                await asyncio.sleep(2)  # Pause for a while before retrying

    def parse_event_payload(self, msg) -> Tuple[List[ZetaEvent], int]:
        slot = int(msg[0].result.context.slot)
        logs = cast(list[str], msg[0].result.value.logs)  # type: ignore
        parsed: list[Event] = []
        self.exchange._event_parser.parse_logs(logs, lambda evt: parsed.append(evt))
        events = []
        for event in parsed:
            if event.name.startswith(PlaceOrderEvent.__name__):
                place_order_event = PlaceOrderEvent.from_event(event)
                if place_order_event.margin_account == self._margin_account_address:
                    events.append(place_order_event)
            elif event.name.startswith(OrderCompleteEvent.__name__):
                order_complete_event = OrderCompleteEvent.from_event(event)
                # Ignore fills
                if (
                    order_complete_event.order_complete_type == OrderCompleteType.Cancel
                    or order_complete_event.order_complete_type == OrderCompleteType.Booted
                ):
                    cancel_event = CancelOrderEvent.from_order_complete_event(order_complete_event)
                    if cancel_event.margin_account == self._margin_account_address:
                        events.append(cancel_event)
            elif event.name.startswith(TradeEvent.__name__):
                trade_event = TradeEvent.from_event(event)
                if trade_event.margin_account == self._margin_account_address:
                    events.append(trade_event)
            elif event.name.startswith(LiquidationEvent.__name__):
                liquidation_event = LiquidationEvent.from_event(event)
                if liquidation_event.liquidatee_margin_account == self._margin_account_address:
                    events.append(liquidation_event)
            else:
                pass

        return events, slot

    async def subscribe_transactions(
        self,
        commitment: Optional[Commitment] = None,
        max_retries: int = 3,
    ) -> AsyncIterator[Tuple[List[ZetaEnrichedEvent], int]]:
        """
        This method is used to subscribe to transactions.

        Args:
            commitment (Optional[Commitment], optional): The commitment level to use for the subscription. Defaults to None.
            max_retries (int, optional): The maximum number of retries for the subscription in case of failure. Defaults to 3.

        Yields:
            List[ZetaEvent]: A list of ZetaEvents that are yielded as they are received.

        Warning:
            This method is experimental and requires a Triton RPC node.
        """
        # self._logger.warning("This method is experimental and requires a Triton RPC node.")
        commitment = commitment or self.connection.commitment
        retries = max_retries
        while retries > 0:
            try:
                async with websockets.connect(self.ws_endpoint + "/whirligig") as ws:  # type: ignore
                    transaction_subscribe = request(
                        "transactionSubscribe",
                        params=[
                            {
                                "mentions": [str(self._margin_account_address)],
                                "failed": False,
                                "vote": False,
                            },
                            {
                                "commitment": str(commitment),
                            },
                        ],
                    )

                    await ws.send(json.dumps(transaction_subscribe))
                    first_resp = await ws.recv()
                    subscription_id = cast(int, first_resp)

                    async for msg in ws:
                        try:
                            events, slot = self.parse_transaction_payload(msg)
                            if len(events) > 0:
                                yield events, slot

                        except Exception:
                            self._logger.error(f"Error processing transaction data: {traceback.format_exc()}")
                            break

                    request(
                        "transactionUnsubscribe",
                        params=[subscription_id],
                    )
                    await ws.send(json.dumps(transaction_subscribe))

            except asyncio.CancelledError:
                self._logger.info("WebSocket subscription task cancelled.")
                break
            except websockets.exceptions.ConnectionClosed:
                self._logger.error("WebSocket connection closed unexpectedly. Attempting to reconnect...")
            except Exception as e:
                self._logger.error(f"Error subscribing to {self.__class__.__name__}: {e}. Retrying...")
                retries -= 1
                if retries <= 0:
                    self._logger.error("Max retries reached. Unable to subscribe to transactions.")
                    break
                await asyncio.sleep(2)  # Pause for a while before retrying

    def parse_transaction_payload(self, msg) -> Tuple[List[ZetaEnrichedEvent], int]:
        json_msg = json.loads(msg)
        slot = int(json_msg["params"]["result"]["context"]["slot"])
        tx_value = json_msg["params"]["result"]["value"]
        log_messages = tx_value["meta"]["logMessages"]
        message = tx_value["transaction"]["message"]

        if isinstance(message[0], int) or "instructions" not in message[0]:
            message_indexed = message[1]
        else:
            message_indexed = message[0]

        ixs = message_indexed["instructions"][1:]
        ix_args: list[Optional[Container]] = []
        ix_names: list[Optional[str]] = []
        events_to_return = []

        for ix in ixs:
            acc_keys_raw = message_indexed["accountKeys"][1:]
            account_keys = [str(based58.b58encode(bytes(a)), encoding="utf-8") for a in acc_keys_raw]
            loaded_addresses = tx_value["meta"]["loadedAddresses"]
            loaded_addresses_list = account_keys + loaded_addresses["writable"] + loaded_addresses["readonly"]
            owner_address = loaded_addresses_list[ix["programIdIndex"]]

            if owner_address != str(constants.ZETA_PID[self.network]):
                ix_args.append(None)
                ix_names.append(None)
                continue
            data = self.exchange.program.coder.instruction.parse(bytes(ix["data"][1:]))
            ix_args.append(data.data)
            ix_names.append(data.name)

        # Split log_messages every time we see "invoke [1]"
        split_log_messages = []
        split_indices = []
        for i in range(len(log_messages)):
            if log_messages[i] == "Log truncated":
                raise Exception("Logs truncated, missing event data")
            if log_messages[i].endswith("invoke [1]"):
                split_indices.append(i)

        split_log_messages = [log_messages[i:j] for i, j in zip([0] + split_indices, split_indices + [None])]
        if len(split_log_messages) > 0:
            split_log_messages = split_log_messages[1:]

        if len(ix_args) != len(split_log_messages) or len(ix_names) != len(split_log_messages):
            raise Exception("Mismatched transaction info lengths")

        # For each individual instruction, find the ix name and the events
        for i in range(len(split_log_messages)):
            # First log line will always be "...invoke [1]", second will be "Program log: Instruction: <ix_name>"
            ix_name = ix_names[i]
            ix_arg = ix_args[i]

            if ix_name is None or ix_arg is None:
                continue

            chunk = split_log_messages[i]
            events: list[Event] = []
            self.exchange._event_parser.parse_logs(chunk, lambda evt: events.append(evt))

            # Depending on the instruction and event we can get different data from the args
            for event in events:
                # Skip event that aren't for our account but mention our account
                # eg if we do a taker trade, we want to skip the maker crank events
                if str(event.data.margin_account) != str(self._margin_account_address):
                    continue

                if ix_name.startswith("place_perp_order"):
                    if event.name.startswith(TradeEvent.__name__):
                        events_to_return.append(TradeEvent.from_event(event))  # Taker fill
                    elif event.name.startswith(PlaceOrderEvent.__name__):
                        events_to_return.append(PlaceOrderEventWithArgs.from_event_and_args(event, ix_arg))

                elif ix_name.startswith("crank_event_queue"):
                    if event.name.startswith(TradeEvent.__name__):
                        events_to_return.append(TradeEvent.from_event(event))  # Maker fill\

                elif ix_name.startswith("cancel_"):
                    if event.name.startswith(OrderCompleteEvent.__name__):
                        order_complete_event = OrderCompleteEvent.from_event(event)
                        if (
                            order_complete_event.order_complete_type == OrderCompleteType.Cancel
                            or order_complete_event.order_complete_type == OrderCompleteType.Booted
                        ):
                            events_to_return.append(CancelOrderEvent.from_order_complete_event(order_complete_event))

        return events_to_return, slot

    # Instructions

    async def deposit(self, amount: float, subaccount_index: int = 0):
        ixs = []
        if not await self._check_margin_account_manager_exists():
            self._logger.info("User has no cross-margin account manager, creating one...")
            ixs.append(self._init_margin_account_manager_ix())
        # Create margin account if user doesn't have one
        if not await self._check_margin_account_exists():
            self._logger.info("User has no cross-margin account, creating one...")
            ixs.append(self._init_margin_account_ix(subaccount_index))
        # Check they have an existing USDC account
        if await self._check_user_usdc_account_exists():
            ixs.append(self._deposit_ix(amount))
        else:
            raise Exception("User has no USDC, cannot deposit to margin account")

        self._logger.info(f"Depositing {amount} USDC to margin account")
        return await self._send_versioned_transaction(ixs)

    def _init_margin_account_manager_ix(self) -> Instruction:
        if self._margin_account_manager_address is None:
            raise Exception("Margin account manager address not loaded, cannot deposit")
        return initialize_cross_margin_account_manager(
            {
                "cross_margin_account_manager": self._margin_account_manager_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "zeta_program": self.exchange.program_id,
            },
            self.exchange.program_id,
        )

    def _init_margin_account_ix(self, subaccount_index: int = 0) -> Instruction:
        if self._margin_account_address is None or self._margin_account_manager_address is None:
            raise Exception("Margin account address not loaded, cannot deposit")
        return initialize_cross_margin_account(
            {"subaccount_index": subaccount_index},
            {
                "cross_margin_account": self._margin_account_address,
                "cross_margin_account_manager": self._margin_account_manager_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "zeta_program": self.exchange.program_id,
            },
            self.exchange.program_id,
        )

    def _deposit_ix(self, amount: float) -> Instruction:
        if self._user_usdc_address is None or self._margin_account_address is None:
            raise Exception("User USDC address not loaded, cannot deposit")
        return deposit_v2(
            {"amount": utils.convert_decimal_to_fixed_int(amount)},
            {
                "margin_account": self._margin_account_address,
                "vault": self._combined_vault_address,
                "user_token_account": self._user_usdc_address,
                "socialized_loss_account": self._combined_socialized_loss_address,
                "authority": self.provider.wallet.public_key,
                "state": self.exchange._state_address,
                "pricing": self.exchange._pricing_address,
            },
            self.exchange.program_id,
        )

    # TODO: withdraw (and optionally close)
    async def withdraw(self):
        raise NotImplementedError

    def _init_open_orders_ix(self, asset: Asset) -> Instruction:
        if asset not in self.exchange.assets:
            raise Exception(f"Asset {asset.name} not loaded into client, cannot initialize open orders")
        if self._open_orders_addresses is None or self._margin_account_address is None:
            raise Exception("Open orders address not loaded, cannot place order")
        return initialize_open_orders_v3(
            {"asset": asset.to_program_type()},
            {
                "state": self.exchange._state_address,
                "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                "open_orders": self._open_orders_addresses[asset],
                "cross_margin_account": self._margin_account_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "market": self.exchange.markets[asset].address,
                "serum_authority": self.exchange._serum_authority_address,
                "open_orders_map": pda.get_open_orders_map_address(
                    self.exchange.program_id, self._open_orders_addresses[asset]
                ),
            },
            self.exchange.program_id,
        )

    def _place_order_ix(
        self,
        asset: Asset,
        price: float,
        size: float,
        side: Side,
        order_opts: Optional[OrderOptions] = None,
    ) -> Instruction:
        if order_opts is None:
            order_opts = OrderOptions()

        unix_timestamp = int(time.time())
        tif_offset = (
            utils.get_tif_offset(
                order_opts.expiry_ts,
                self.exchange.markets[asset]._market_state.epoch_length,
                unix_timestamp,  # self.exchange.clock.account.unix_timestamp,
            )
            if order_opts.expiry_ts
            else None
        )

        if asset not in self.exchange.assets:
            raise Exception(f"Asset {asset.name} not loaded into client, cannot place order")
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot place order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot place order")

        return place_perp_order_v3(
            {
                "price": utils.convert_decimal_to_fixed_int(price),
                "size": utils.convert_decimal_to_fixed_lot(size),
                "side": side.to_program_type(),
                "order_type": order_opts.order_type.to_program_type(),
                "client_order_id": order_opts.client_order_id,
                "tif_offset": tif_offset,
                "tag": order_opts.tag,
                "asset": asset.to_program_type(),
            },
            {
                "state": self.exchange._state_address,
                "pricing": self.exchange._pricing_address,
                "margin_account": self._margin_account_address,
                "authority": self.provider.wallet.public_key,
                "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                "serum_authority": self.exchange._serum_authority_address,
                "open_orders": self._open_orders_addresses[asset],
                "market_accounts": {
                    "market": self.exchange.markets[asset].address,
                    "request_queue": self.exchange.markets[asset]._market_state.request_queue,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "coin_vault": self.exchange.markets[asset]._market_state.base_vault,
                    "pc_vault": self.exchange.markets[asset]._market_state.quote_vault,
                    "order_payer_token_account": self.exchange.markets[asset]._quote_zeta_vault_address
                    if side == Side.Bid
                    else self.exchange.markets[asset]._base_zeta_vault_address,
                    "coin_wallet": self.exchange.markets[asset]._base_zeta_vault_address,
                    "pc_wallet": self.exchange.markets[asset]._quote_zeta_vault_address,
                },
                "oracle": self.exchange.pricing.oracles[asset.to_index()],
                "oracle_backup_feed": self.exchange.pricing.oracle_backup_feeds[asset.to_index()],
                "oracle_backup_program": constants.CHAINLINK_PID,
                "market_mint": self.exchange.markets[asset]._market_state.quote_mint
                if side == Side.Bid
                else self.exchange.markets[asset]._market_state.base_mint,
                "mint_authority": self.exchange._mint_authority_address,
                "perp_sync_queue": self.exchange.pricing.perp_sync_queues[asset.to_index()],
            },
            self.exchange.program_id,
        )

    async def cancel_order(self, asset: Asset, order_id: int, side: Side):
        ixs = [self._cancel_order_ix(asset, order_id, side)]
        self._logger.info(f"Cancelling order {order_id} for {asset}")
        return await self._send_versioned_transaction(ixs)

    def _cancel_order_ix(self, asset: Asset, order_id: int, side: Side) -> Instruction:
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot cancel order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot cancel order")
        return cancel_order(
            {"side": side.to_program_type(), "order_id": order_id, "asset": asset.to_program_type()},
            {
                "authority": self.provider.wallet.public_key,
                "cancel_accounts": {
                    "state": self.exchange._state_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market": self.exchange.markets[asset].address,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                },
            },
            self.exchange.program_id,
        )

    # TODO: cancelorderbyclientorderid

    def _cancel_orders_for_market_ix(self, asset: Asset) -> Instruction:
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot cancel orders")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot cancel orders")
        return cancel_all_market_orders(
            {"asset": asset.to_program_type()},
            {
                "authority": self.provider.wallet.public_key,
                "cancel_accounts": {
                    "state": self.exchange._state_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market": self.exchange.markets[asset].address,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                },
            },
            self.exchange.program_id,
        )

    async def cancel_orders_for_market(self, asset: Asset):
        ixs = [self._cancel_orders_for_market_ix(asset)]
        self._logger.info(f"Cancelling all orders for {asset}")
        return await self._send_versioned_transaction(ixs)

    async def place_orders_for_market(
        self,
        asset: Asset,
        orders: list[OrderArgs],
        pre_instructions: Optional[list[Instruction]] = None,
        post_instructions: Optional[list[Instruction]] = None,
    ):
        # TODO: warn about log truncation above 10 orders
        ixs = []
        if not await self._check_open_orders_account_exists(asset):
            self._logger.info("User has no open orders account, creating one...")
            ixs.append(self._init_open_orders_ix(asset))

        if pre_instructions is not None:
            ixs.extend(pre_instructions)
        for order in orders:
            ixs.append(self._place_order_ix(asset, order.price, order.size, order.side, order.order_opts))
        if post_instructions is not None:
            ixs.extend(post_instructions)
        self._logger.info(f"Placing {len(orders)} orders for {asset}")

        return await self._send_versioned_transaction(ixs)

    async def replace_orders_for_market(
        self,
        asset: Asset,
        orders: list[OrderArgs],
    ):
        return await self.place_orders_for_market(
            asset, orders, pre_instructions=[self._cancel_orders_for_market_ix(asset)]
        )

    # TODO: liquidate
    async def liquidate(self):
        raise NotImplementedError

    async def _send_versioned_transaction(self, ixs: list[Instruction]):
        # Prefetch blockhash, using cache if available
        if self.connection.blockhash_cache:
            try:
                recent_blockhash = self.connection.blockhash_cache.get()
                last_valid_block_height = None
                self._logger.debug(f"Blockhash cache hit, using cached blockhash: {recent_blockhash}")
            except ValueError:
                blockhash_resp = await self.connection.get_latest_blockhash(constants.BLOCKHASH_COMMITMENT)
                recent_blockhash = self.connection._process_blockhash_resp(blockhash_resp, used_immediately=True)
                last_valid_block_height = blockhash_resp.value.last_valid_block_height
                self._logger.debug(f"Blockhash cache miss, fetched from RPC: {recent_blockhash}")
        else:
            blockhash_resp = await self.connection.get_latest_blockhash(constants.BLOCKHASH_COMMITMENT)
            recent_blockhash = self.connection.parse_recent_blockhash(blockhash_resp)
            last_valid_block_height = blockhash_resp.value.last_valid_block_height
            self._logger.debug(f"Blockhash cache not enabled, fetched from RPC: {recent_blockhash}")

        msg = MessageV0.try_compile(
            self.provider.wallet.public_key, ixs, [constants.ZETA_LUT[self.network]], recent_blockhash
        )
        tx = VersionedTransaction(msg, [self.provider.wallet.payer])
        try:
            opts = self.provider.opts._replace(last_valid_block_height=last_valid_block_height)
            signature = await self.provider.send(tx, opts)
        except RPCException as exc:
            # This won't work on zDEX errors
            # TODO: add ZDEX error parsing
            parsed = from_tx_error(exc)
            self._logger.error(parsed)
            if parsed is not None:
                raise parsed from exc
            raise exc
        return signature



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeReferrerAliasArgs(typing.TypedDict):
    alias: str


layout = borsh.CStruct("alias" / borsh.String)


class InitializeReferrerAliasAccounts(typing.TypedDict):
    referrer: Pubkey
    referrer_alias: Pubkey
    referrer_account: Pubkey


def initialize_referrer_alias(
    args: InitializeReferrerAliasArgs,
    accounts: InitializeReferrerAliasAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["referrer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["referrer_alias"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["referrer_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"C\x1c\x03N\x8cl\x14\x8a"
    encoded_args = layout.build(
        {
            "alias": args["alias"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import order_state, position


class ProductLedgerJSON(typing.TypedDict):
    position: position.PositionJSON
    order_state: order_state.OrderStateJSON


@dataclass
class ProductLedger:
    layout: typing.ClassVar = borsh.CStruct(
        "position" / position.Position.layout,
        "order_state" / order_state.OrderState.layout,
    )
    position: position.Position
    order_state: order_state.OrderState

    @classmethod
    def from_decoded(cls, obj: Container) -> "ProductLedger":
        return cls(
            position=position.Position.from_decoded(obj.position),
            order_state=order_state.OrderState.from_decoded(obj.order_state),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "position": self.position.to_encodable(),
            "order_state": self.order_state.to_encodable(),
        }

    def to_json(self) -> ProductLedgerJSON:
        return {
            "position": self.position.to_json(),
            "order_state": self.order_state.to_json(),
        }

    @classmethod
    def from_json(cls, obj: ProductLedgerJSON) -> "ProductLedger":
        return cls(
            position=position.Position.from_json(obj["position"]),
            order_state=order_state.OrderState.from_json(obj["order_state"]),
        )



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from ..program_id import PROGRAM_ID


class ReferUserAccounts(typing.TypedDict):
    user: Pubkey
    referrer_account: Pubkey
    referral_account: Pubkey


def refer_user(
    accounts: ReferUserAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["user"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["referrer_account"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["referral_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"9r\x9a\xcc\xb9!\xaa\x88"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from construct import Container, Padding, Switch
from solders.pubkey import Pubkey


@dataclass
class SlabHeader:
    layout: typing.ClassVar = borsh.CStruct(
        "bump_index" / borsh.U32,
        "padding_1" / Padding(4),
        "free_list_length" / borsh.U32,
        "padding_2" / Padding(4),
        "free_list_head" / borsh.U32,
        "root" / borsh.U32,
        "leaf_count" / borsh.U32,
        "padding_3" / Padding(4),
    )
    bump_index: int
    free_list_length: int
    free_list_head: int
    root: int
    leaf_count: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SlabHeader":
        return cls(
            bump_index=obj.bump_index,
            free_list_length=obj.free_list_length,
            free_list_head=obj.free_list_head,
            root=obj.root,
            leaf_count=obj.leaf_count,
        )


# Different node types, we pad it all to size of 68 bytes.
UNINTIALIZED = borsh.CStruct("padding" / borsh.U8[68])
FREE_NODE = borsh.CStruct("next" / borsh.U32, "padding" / borsh.U8[64])
LAST_FREE_NODE = borsh.CStruct("padding" / borsh.U8[68])

# Used as dummy value for SlabNode#next.
NONE_NEXT = -1


@dataclass
class SlabInnerNode:
    layout: typing.ClassVar = borsh.CStruct(
        "prefix_len" / borsh.U32, "key" / borsh.U128, "children" / borsh.U32[2], "padding" / borsh.U8[40]
    )
    prefix_len: int
    key: int
    children: list[int]
    is_initialized: bool = True
    next: int = NONE_NEXT

    @classmethod
    def from_decoded(cls, obj: Container) -> "SlabInnerNode":
        return cls(
            prefix_len=obj.prefix_len,
            key=obj.key,
            children=obj.children,
        )


@dataclass
class SlabLeafNode:
    layout: typing.ClassVar = borsh.CStruct(
        "owner_slot" / borsh.U8,
        "fee_tier" / borsh.U8,
        "tif_offset" / borsh.U16,
        "key" / borsh.U128,
        "owner" / BorshPubkey,
        "quantity" / borsh.U64,
        "client_order_id" / borsh.U64,
    )
    owner_slot: int
    fee_tier: int
    tif_offset: int
    key: int
    owner: Pubkey
    quantity: int
    client_order_id: int
    is_initialized: bool = True
    next: int = NONE_NEXT

    @classmethod
    def from_decoded(cls, obj: Container) -> "SlabLeafNode":
        return cls(
            owner_slot=obj.owner_slot,
            fee_tier=obj.fee_tier,
            tif_offset=obj.tif_offset,
            key=obj.key,
            owner=obj.owner,
            quantity=obj.quantity,
            client_order_id=obj.client_order_id,
        )


@dataclass
class SlabNode:
    layout: typing.ClassVar = borsh.CStruct(
        "tag" / borsh.U32,
        "node"
        / Switch(
            lambda this: this.tag,
            {
                0: UNINTIALIZED,
                1: SlabInnerNode.layout,
                2: SlabLeafNode.layout,
                3: FREE_NODE,
                4: LAST_FREE_NODE,
            },
        ),
    )
    is_initialized: bool
    next: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SlabNode" | SlabInnerNode | SlabLeafNode:
        if obj.tag == 0:
            return cls(is_initialized=False, next=NONE_NEXT)
        elif obj.tag == 1:
            return SlabInnerNode.from_decoded(obj.node)
        elif obj.tag == 2:
            return SlabLeafNode.from_decoded(obj.node)
        elif obj.tag == 3:
            return cls(is_initialized=True, next=obj.node.next)
        elif obj.tag == 4:
            return cls(is_initialized=True, next=NONE_NEXT)
        else:
            raise RuntimeError("Invalid tag!")


@dataclass
class Slab:
    layout: typing.ClassVar = borsh.CStruct(
        "header" / SlabHeader.layout, "nodes" / SlabNode.layout[lambda this: this.header.bump_index]
    )
    header: SlabHeader
    nodes: list[SlabNode | SlabInnerNode | SlabLeafNode]

    @classmethod
    def from_decoded(cls, obj: Container) -> "Slab":
        return cls(
            header=SlabHeader.from_decoded(obj.header),
            nodes=list(
                map(
                    lambda item: SlabNode.from_decoded(item),
                    obj.nodes,
                )
            ),
        )

    def __iter__(self) -> typing.Iterable[SlabLeafNode]:
        return self.items(False)

    def items(self, descending=False) -> typing.Iterable[SlabLeafNode]:
        """Depth first traversal of the Binary Tree.
        Parameter descending decides if the price should descending or not.
        """
        if self.header.leaf_count == 0:
            return
        stack = [self.header.root]
        while stack:
            index = stack.pop()
            node: SlabNode | SlabInnerNode | SlabLeafNode = self.nodes[index]
            if isinstance(node, SlabLeafNode):
                yield node
            elif isinstance(node, SlabInnerNode):
                if descending:
                    stack.append(node.children[0])
                    stack.append(node.children[1])
                else:
                    stack.append(node.children[1])
                    stack.append(node.children[0])
            else:
                raise RuntimeError("Neither of leaf node or tree node!")



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class InitialJSON(typing.TypedDict):
    kind: typing.Literal["Initial"]


class MaintenanceJSON(typing.TypedDict):
    kind: typing.Literal["Maintenance"]


class MaintenanceIncludingOrdersJSON(typing.TypedDict):
    kind: typing.Literal["MaintenanceIncludingOrders"]


class MarketMakerConcessionJSON(typing.TypedDict):
    kind: typing.Literal["MarketMakerConcession"]


@dataclass
class Initial:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Initial"

    @classmethod
    def to_json(cls) -> InitialJSON:
        return InitialJSON(
            kind="Initial",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Initial": {},
        }


@dataclass
class Maintenance:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Maintenance"

    @classmethod
    def to_json(cls) -> MaintenanceJSON:
        return MaintenanceJSON(
            kind="Maintenance",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Maintenance": {},
        }


@dataclass
class MaintenanceIncludingOrders:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "MaintenanceIncludingOrders"

    @classmethod
    def to_json(cls) -> MaintenanceIncludingOrdersJSON:
        return MaintenanceIncludingOrdersJSON(
            kind="MaintenanceIncludingOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "MaintenanceIncludingOrders": {},
        }


@dataclass
class MarketMakerConcession:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "MarketMakerConcession"

    @classmethod
    def to_json(cls) -> MarketMakerConcessionJSON:
        return MarketMakerConcessionJSON(
            kind="MarketMakerConcession",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "MarketMakerConcession": {},
        }


MarginRequirementKind = typing.Union[Initial, Maintenance, MaintenanceIncludingOrders, MarketMakerConcession]
MarginRequirementJSON = typing.Union[
    InitialJSON,
    MaintenanceJSON,
    MaintenanceIncludingOrdersJSON,
    MarketMakerConcessionJSON,
]


def from_decoded(obj: dict) -> MarginRequirementKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Initial" in obj:
        return Initial()
    if "Maintenance" in obj:
        return Maintenance()
    if "MaintenanceIncludingOrders" in obj:
        return MaintenanceIncludingOrders()
    if "MarketMakerConcession" in obj:
        return MarketMakerConcession()
    raise ValueError("Invalid enum object")


def from_json(obj: MarginRequirementJSON) -> MarginRequirementKind:
    if obj["kind"] == "Initial":
        return Initial()
    if obj["kind"] == "Maintenance":
        return Maintenance()
    if obj["kind"] == "MaintenanceIncludingOrders":
        return MaintenanceIncludingOrders()
    if obj["kind"] == "MarketMakerConcession":
        return MarketMakerConcession()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Initial" / borsh.CStruct(),
    "Maintenance" / borsh.CStruct(),
    "MaintenanceIncludingOrders" / borsh.CStruct(),
    "MarketMakerConcession" / borsh.CStruct(),
)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class CloseOpenOrdersV3Args(typing.TypedDict):
    map_nonce: int
    asset: types.asset.AssetKind


layout = borsh.CStruct("map_nonce" / borsh.U8, "asset" / types.asset.layout)


class CloseOpenOrdersV3Accounts(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    dex_program: Pubkey
    open_orders: Pubkey
    cross_margin_account: Pubkey
    authority: Pubkey
    market: Pubkey
    serum_authority: Pubkey
    open_orders_map: Pubkey


def close_open_orders_v3(
    args: CloseOpenOrdersV3Args,
    accounts: CloseOpenOrdersV3Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["cross_margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders_map"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcf\x0f\xc6J\xc5\xe4\xb0\x1e"
    encoded_args = layout.build(
        {
            "map_nonce": args["map_nonce"],
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class InitializedJSON(typing.TypedDict):
    kind: typing.Literal["Initialized"]


class LiveJSON(typing.TypedDict):
    kind: typing.Literal["Live"]


class ExpiredJSON(typing.TypedDict):
    kind: typing.Literal["Expired"]


class ExpiredDirtyJSON(typing.TypedDict):
    kind: typing.Literal["ExpiredDirty"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Initialized:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Initialized"

    @classmethod
    def to_json(cls) -> InitializedJSON:
        return InitializedJSON(
            kind="Initialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Initialized": {},
        }


@dataclass
class Live:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Live"

    @classmethod
    def to_json(cls) -> LiveJSON:
        return LiveJSON(
            kind="Live",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Live": {},
        }


@dataclass
class Expired:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Expired"

    @classmethod
    def to_json(cls) -> ExpiredJSON:
        return ExpiredJSON(
            kind="Expired",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Expired": {},
        }


@dataclass
class ExpiredDirty:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "ExpiredDirty"

    @classmethod
    def to_json(cls) -> ExpiredDirtyJSON:
        return ExpiredDirtyJSON(
            kind="ExpiredDirty",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ExpiredDirty": {},
        }


ExpirySeriesStatusKind = typing.Union[Uninitialized, Initialized, Live, Expired, ExpiredDirty]
ExpirySeriesStatusJSON = typing.Union[UninitializedJSON, InitializedJSON, LiveJSON, ExpiredJSON, ExpiredDirtyJSON]


def from_decoded(obj: dict) -> ExpirySeriesStatusKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Initialized" in obj:
        return Initialized()
    if "Live" in obj:
        return Live()
    if "Expired" in obj:
        return Expired()
    if "ExpiredDirty" in obj:
        return ExpiredDirty()
    raise ValueError("Invalid enum object")


def from_json(obj: ExpirySeriesStatusJSON) -> ExpirySeriesStatusKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Initialized":
        return Initialized()
    if obj["kind"] == "Live":
        return Live()
    if obj["kind"] == "Expired":
        return Expired()
    if obj["kind"] == "ExpiredDirty":
        return ExpiredDirty()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "Initialized" / borsh.CStruct(),
    "Live" / borsh.CStruct(),
    "Expired" / borsh.CStruct(),
    "ExpiredDirty" / borsh.CStruct(),
)



from dataclasses import dataclass, field
from enum import Enum, IntEnum
from typing import Optional

from solders.hash import Hash
from solders.pubkey import Pubkey

from zetamarkets_py.zeta_client.types import asset, order_type, side


class Asset(Enum):
    SOL = "SOL"
    BTC = "BTC"
    ETH = "ETH"
    APT = "APT"
    ARB = "ARB"

    def to_index(self):
        members = list(self.__class__.__members__.values())
        return members.index(self)

    @classmethod
    def from_index(cls, index: int):
        members = list(Asset)
        return members[index]

    def to_program_type(self):
        return asset.from_decoded({self.name: self.to_index()})

    @staticmethod
    def all():
        return list(Asset)

    def __str__(self) -> str:
        return self.name


class Network(Enum):
    DEVNET = "devnet"
    TESTNET = "testnet"
    MAINNET = "mainnet_beta"

    def __str__(self) -> str:
        return self.name


class OrderType(IntEnum):
    Limit = 0
    PostOnly = 1
    FillOrKill = 2
    ImmediateOrCancel = 3
    PostOnlySlide = 4

    def to_program_type(self):
        return order_type.from_decoded({self.name: self.value})

    @classmethod
    def from_index(cls, index: int):
        members = list(OrderType)
        return members[index]

    def __str__(self) -> str:
        return self.name


class Side(Enum):
    """Side of the orderbook to trade."""

    Uninitialized = 0
    """"""
    Bid = 1
    """"""
    Ask = 2
    """"""

    def to_program_type(self) -> side.SideKind:
        return side.from_decoded({self.name: self.value})

    @classmethod
    def from_index(cls, index: int):
        members = list(Side)
        return members[index]

    def __str__(self) -> str:
        return self.name


class SelfTradeBehavior(Enum):
    DecrementTake = 0
    CancelProvide = 1
    AbortTransaction = 2

    def __str__(self) -> str:
        return self.name


class OrderCompleteType(Enum):
    Cancel = 0
    Fill = 1
    Booted = 2

    @classmethod
    def from_index(cls, index: int):
        members = list(OrderCompleteType)
        return members[index]

    def __str__(self) -> str:
        return self.name


@dataclass
class Position:
    size: float
    cost_of_trades: float


@dataclass
class TIFOptions:
    expiry_offset: Optional[int] = None
    expiry_ts: Optional[int] = None


@dataclass
class OrderOptions:
    expiry_ts: Optional[int] = None
    client_order_id: Optional[int] = None
    blockhash: Optional[Hash] = None
    order_type: OrderType = OrderType.Limit
    tag: str = "SDK"


@dataclass
class OrderInfo:
    price: float
    size: float


@dataclass
class Order:
    order_id: int
    client_id: int
    open_order_address: Pubkey
    open_order_slot: int
    fee_tier: int
    info: OrderInfo
    side: Side
    tif_offset: int


@dataclass
class FilledOrder:
    order_id: int
    side: Side
    price: float
    size: float
    fee_cost: int


@dataclass
class OrderArgs:
    price: float
    size: float
    side: Side
    order_opts: OrderOptions = field(default_factory=OrderOptions)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdatePricingParametersArgsJSON(typing.TypedDict):
    option_trade_normalizer: int
    future_trade_normalizer: int
    max_volatility_retreat: int
    max_interest_retreat: int
    min_delta: int
    max_delta: int
    min_interest_rate: int
    max_interest_rate: int
    min_volatility: int
    max_volatility: int


@dataclass
class UpdatePricingParametersArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "option_trade_normalizer" / borsh.U64,
        "future_trade_normalizer" / borsh.U64,
        "max_volatility_retreat" / borsh.U64,
        "max_interest_retreat" / borsh.U64,
        "min_delta" / borsh.U64,
        "max_delta" / borsh.U64,
        "min_interest_rate" / borsh.I64,
        "max_interest_rate" / borsh.I64,
        "min_volatility" / borsh.U64,
        "max_volatility" / borsh.U64,
    )
    option_trade_normalizer: int
    future_trade_normalizer: int
    max_volatility_retreat: int
    max_interest_retreat: int
    min_delta: int
    max_delta: int
    min_interest_rate: int
    max_interest_rate: int
    min_volatility: int
    max_volatility: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdatePricingParametersArgs":
        return cls(
            option_trade_normalizer=obj.option_trade_normalizer,
            future_trade_normalizer=obj.future_trade_normalizer,
            max_volatility_retreat=obj.max_volatility_retreat,
            max_interest_retreat=obj.max_interest_retreat,
            min_delta=obj.min_delta,
            max_delta=obj.max_delta,
            min_interest_rate=obj.min_interest_rate,
            max_interest_rate=obj.max_interest_rate,
            min_volatility=obj.min_volatility,
            max_volatility=obj.max_volatility,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "option_trade_normalizer": self.option_trade_normalizer,
            "future_trade_normalizer": self.future_trade_normalizer,
            "max_volatility_retreat": self.max_volatility_retreat,
            "max_interest_retreat": self.max_interest_retreat,
            "min_delta": self.min_delta,
            "max_delta": self.max_delta,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
        }

    def to_json(self) -> UpdatePricingParametersArgsJSON:
        return {
            "option_trade_normalizer": self.option_trade_normalizer,
            "future_trade_normalizer": self.future_trade_normalizer,
            "max_volatility_retreat": self.max_volatility_retreat,
            "max_interest_retreat": self.max_interest_retreat,
            "min_delta": self.min_delta,
            "max_delta": self.max_delta,
            "min_interest_rate": self.min_interest_rate,
            "max_interest_rate": self.max_interest_rate,
            "min_volatility": self.min_volatility,
            "max_volatility": self.max_volatility,
        }

    @classmethod
    def from_json(cls, obj: UpdatePricingParametersArgsJSON) -> "UpdatePricingParametersArgs":
        return cls(
            option_trade_normalizer=obj["option_trade_normalizer"],
            future_trade_normalizer=obj["future_trade_normalizer"],
            max_volatility_retreat=obj["max_volatility_retreat"],
            max_interest_retreat=obj["max_interest_retreat"],
            min_delta=obj["min_delta"],
            max_delta=obj["max_delta"],
            min_interest_rate=obj["min_interest_rate"],
            max_interest_rate=obj["max_interest_rate"],
            min_volatility=obj["min_volatility"],
            max_volatility=obj["max_volatility"],
        )



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class CrossMarginAccountJSON(typing.TypedDict):
    authority: str
    delegated_pubkey: str
    balance: int
    subaccount_index: int
    nonce: int
    force_cancel_flag: bool
    account_type: types.margin_account_type.MarginAccountTypeJSON
    open_orders_nonces: list[int]
    open_orders_nonces_padding: list[int]
    rebalance_amount: int
    last_funding_deltas: list[types.anchor_decimal.AnchorDecimalJSON]
    last_funding_deltas_padding: list[types.anchor_decimal.AnchorDecimalJSON]
    product_ledgers: list[types.product_ledger.ProductLedgerJSON]
    product_ledgers_padding: list[types.product_ledger.ProductLedgerJSON]
    padding: list[int]


@dataclass
class CrossMarginAccount:
    discriminator: typing.ClassVar = b"\xf2^\x8e\x83#\xf4\x93\x1c"
    layout: typing.ClassVar = borsh.CStruct(
        "authority" / BorshPubkey,
        "delegated_pubkey" / BorshPubkey,
        "balance" / borsh.U64,
        "subaccount_index" / borsh.U8,
        "nonce" / borsh.U8,
        "force_cancel_flag" / borsh.Bool,
        "account_type" / types.margin_account_type.layout,
        "open_orders_nonces" / borsh.U8[5],
        "open_orders_nonces_padding" / borsh.U8[20],
        "rebalance_amount" / borsh.I64,
        "last_funding_deltas" / types.anchor_decimal.AnchorDecimal.layout[5],
        "last_funding_deltas_padding" / types.anchor_decimal.AnchorDecimal.layout[20],
        "product_ledgers" / types.product_ledger.ProductLedger.layout[5],
        "product_ledgers_padding" / types.product_ledger.ProductLedger.layout[20],
        "padding" / borsh.U8[2000],
    )
    authority: Pubkey
    delegated_pubkey: Pubkey
    balance: int
    subaccount_index: int
    nonce: int
    force_cancel_flag: bool
    account_type: types.margin_account_type.MarginAccountTypeKind
    open_orders_nonces: list[int]
    open_orders_nonces_padding: list[int]
    rebalance_amount: int
    last_funding_deltas: list[types.anchor_decimal.AnchorDecimal]
    last_funding_deltas_padding: list[types.anchor_decimal.AnchorDecimal]
    product_ledgers: list[types.product_ledger.ProductLedger]
    product_ledgers_padding: list[types.product_ledger.ProductLedger]
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["CrossMarginAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> list[typing.Optional["CrossMarginAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: list[typing.Optional["CrossMarginAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "CrossMarginAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = CrossMarginAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            authority=dec.authority,
            delegated_pubkey=dec.delegated_pubkey,
            balance=dec.balance,
            subaccount_index=dec.subaccount_index,
            nonce=dec.nonce,
            force_cancel_flag=dec.force_cancel_flag,
            account_type=types.margin_account_type.from_decoded(dec.account_type),
            open_orders_nonces=dec.open_orders_nonces,
            open_orders_nonces_padding=dec.open_orders_nonces_padding,
            rebalance_amount=dec.rebalance_amount,
            last_funding_deltas=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.last_funding_deltas,
                )
            ),
            last_funding_deltas_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_decoded(item),
                    dec.last_funding_deltas_padding,
                )
            ),
            product_ledgers=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_decoded(item),
                    dec.product_ledgers,
                )
            ),
            product_ledgers_padding=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_decoded(item),
                    dec.product_ledgers_padding,
                )
            ),
            padding=dec.padding,
        )

    def to_json(self) -> CrossMarginAccountJSON:
        return {
            "authority": str(self.authority),
            "delegated_pubkey": str(self.delegated_pubkey),
            "balance": self.balance,
            "subaccount_index": self.subaccount_index,
            "nonce": self.nonce,
            "force_cancel_flag": self.force_cancel_flag,
            "account_type": self.account_type.to_json(),
            "open_orders_nonces": self.open_orders_nonces,
            "open_orders_nonces_padding": self.open_orders_nonces_padding,
            "rebalance_amount": self.rebalance_amount,
            "last_funding_deltas": list(map(lambda item: item.to_json(), self.last_funding_deltas)),
            "last_funding_deltas_padding": list(map(lambda item: item.to_json(), self.last_funding_deltas_padding)),
            "product_ledgers": list(map(lambda item: item.to_json(), self.product_ledgers)),
            "product_ledgers_padding": list(map(lambda item: item.to_json(), self.product_ledgers_padding)),
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: CrossMarginAccountJSON) -> "CrossMarginAccount":
        return cls(
            authority=Pubkey.from_string(obj["authority"]),
            delegated_pubkey=Pubkey.from_string(obj["delegated_pubkey"]),
            balance=obj["balance"],
            subaccount_index=obj["subaccount_index"],
            nonce=obj["nonce"],
            force_cancel_flag=obj["force_cancel_flag"],
            account_type=types.margin_account_type.from_json(obj["account_type"]),
            open_orders_nonces=obj["open_orders_nonces"],
            open_orders_nonces_padding=obj["open_orders_nonces_padding"],
            rebalance_amount=obj["rebalance_amount"],
            last_funding_deltas=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["last_funding_deltas"],
                )
            ),
            last_funding_deltas_padding=list(
                map(
                    lambda item: types.anchor_decimal.AnchorDecimal.from_json(item),
                    obj["last_funding_deltas_padding"],
                )
            ),
            product_ledgers=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_json(item),
                    obj["product_ledgers"],
                )
            ),
            product_ledgers_padding=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_json(item),
                    obj["product_ledgers_padding"],
                )
            ),
            padding=obj["padding"],
        )



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


@dataclass
class OrderbookAccount:
    discriminator: typing.ClassVar = b"\x73\x65\x72\x75\x6D"
    layout: typing.ClassVar = borsh.CStruct(
        "account_flags" / types.account_flags.AccountFlags.layout,
        "slab" / types.slab.Slab.layout,
    )
    account_flags: types.account_flags.AccountFlags
    slab: types.slab.Slab

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["OrderbookAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["OrderbookAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["OrderbookAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "OrderbookAccount":
        if data[: len(cls.discriminator)] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = OrderbookAccount.layout.parse(data[len(cls.discriminator) :])
        return cls(
            account_flags=types.account_flags.AccountFlags.from_decoded(dec.account_flags),
            slab=types.slab.Slab.from_decoded(dec.slab),
        )



import typing

from anchorpy.error import ProgramError


class InstructionMissing(ProgramError):
    def __init__(self):
        super().__init__(100, "8 byte instruction identifier not provided")

    code = 100
    name = "InstructionMissing"
    msg = "8 byte instruction identifier not provided"


class InstructionFallbackNotFound(ProgramError):
    def __init__(self):
        super().__init__(101, "Fallback functions are not supported")

    code = 101
    name = "InstructionFallbackNotFound"
    msg = "Fallback functions are not supported"


class InstructionDidNotDeserialize(ProgramError):
    def __init__(self):
        super().__init__(102, "The program could not deserialize the given instruction")

    code = 102
    name = "InstructionDidNotDeserialize"
    msg = "The program could not deserialize the given instruction"


class InstructionDidNotSerialize(ProgramError):
    def __init__(self):
        super().__init__(103, "The program could not serialize the given instruction")

    code = 103
    name = "InstructionDidNotSerialize"
    msg = "The program could not serialize the given instruction"


class IdlInstructionStub(ProgramError):
    def __init__(self):
        super().__init__(1000, "The program was compiled without idl instructions")

    code = 1000
    name = "IdlInstructionStub"
    msg = "The program was compiled without idl instructions"


class IdlInstructionInvalidProgram(ProgramError):
    def __init__(self):
        super().__init__(1001, "The transaction was given an invalid program for the IDL instruction")

    code = 1001
    name = "IdlInstructionInvalidProgram"
    msg = "The transaction was given an invalid program for the IDL instruction"


class ConstraintMut(ProgramError):
    def __init__(self):
        super().__init__(2000, "A mut constraint was violated")

    code = 2000
    name = "ConstraintMut"
    msg = "A mut constraint was violated"


class ConstraintHasOne(ProgramError):
    def __init__(self):
        super().__init__(2001, "A has_one constraint was violated")

    code = 2001
    name = "ConstraintHasOne"
    msg = "A has_one constraint was violated"


class ConstraintSigner(ProgramError):
    def __init__(self):
        super().__init__(2002, "A signer constraint was violated")

    code = 2002
    name = "ConstraintSigner"
    msg = "A signer constraint was violated"


class ConstraintRaw(ProgramError):
    def __init__(self):
        super().__init__(2003, "A raw constraint was violated")

    code = 2003
    name = "ConstraintRaw"
    msg = "A raw constraint was violated"


class ConstraintOwner(ProgramError):
    def __init__(self):
        super().__init__(2004, "An owner constraint was violated")

    code = 2004
    name = "ConstraintOwner"
    msg = "An owner constraint was violated"


class ConstraintRentExempt(ProgramError):
    def __init__(self):
        super().__init__(2005, "A rent exempt constraint was violated")

    code = 2005
    name = "ConstraintRentExempt"
    msg = "A rent exempt constraint was violated"


class ConstraintSeeds(ProgramError):
    def __init__(self):
        super().__init__(2006, "A seeds constraint was violated")

    code = 2006
    name = "ConstraintSeeds"
    msg = "A seeds constraint was violated"


class ConstraintExecutable(ProgramError):
    def __init__(self):
        super().__init__(2007, "An executable constraint was violated")

    code = 2007
    name = "ConstraintExecutable"
    msg = "An executable constraint was violated"


class ConstraintState(ProgramError):
    def __init__(self):
        super().__init__(2008, "A state constraint was violated")

    code = 2008
    name = "ConstraintState"
    msg = "A state constraint was violated"


class ConstraintAssociated(ProgramError):
    def __init__(self):
        super().__init__(2009, "An associated constraint was violated")

    code = 2009
    name = "ConstraintAssociated"
    msg = "An associated constraint was violated"


class ConstraintAssociatedInit(ProgramError):
    def __init__(self):
        super().__init__(2010, "An associated init constraint was violated")

    code = 2010
    name = "ConstraintAssociatedInit"
    msg = "An associated init constraint was violated"


class ConstraintClose(ProgramError):
    def __init__(self):
        super().__init__(2011, "A close constraint was violated")

    code = 2011
    name = "ConstraintClose"
    msg = "A close constraint was violated"


class ConstraintAddress(ProgramError):
    def __init__(self):
        super().__init__(2012, "An address constraint was violated")

    code = 2012
    name = "ConstraintAddress"
    msg = "An address constraint was violated"


class ConstraintZero(ProgramError):
    def __init__(self):
        super().__init__(2013, "Expected zero account discriminant")

    code = 2013
    name = "ConstraintZero"
    msg = "Expected zero account discriminant"


class ConstraintTokenMint(ProgramError):
    def __init__(self):
        super().__init__(2014, "A token mint constraint was violated")

    code = 2014
    name = "ConstraintTokenMint"
    msg = "A token mint constraint was violated"


class ConstraintTokenOwner(ProgramError):
    def __init__(self):
        super().__init__(2015, "A token owner constraint was violated")

    code = 2015
    name = "ConstraintTokenOwner"
    msg = "A token owner constraint was violated"


class ConstraintMintMintAuthority(ProgramError):
    def __init__(self):
        super().__init__(2016, "A mint mint authority constraint was violated")

    code = 2016
    name = "ConstraintMintMintAuthority"
    msg = "A mint mint authority constraint was violated"


class ConstraintMintFreezeAuthority(ProgramError):
    def __init__(self):
        super().__init__(2017, "A mint freeze authority constraint was violated")

    code = 2017
    name = "ConstraintMintFreezeAuthority"
    msg = "A mint freeze authority constraint was violated"


class ConstraintMintDecimals(ProgramError):
    def __init__(self):
        super().__init__(2018, "A mint decimals constraint was violated")

    code = 2018
    name = "ConstraintMintDecimals"
    msg = "A mint decimals constraint was violated"


class ConstraintSpace(ProgramError):
    def __init__(self):
        super().__init__(2019, "A space constraint was violated")

    code = 2019
    name = "ConstraintSpace"
    msg = "A space constraint was violated"


class RequireViolated(ProgramError):
    def __init__(self):
        super().__init__(2500, "A require expression was violated")

    code = 2500
    name = "RequireViolated"
    msg = "A require expression was violated"


class RequireEqViolated(ProgramError):
    def __init__(self):
        super().__init__(2501, "A require_eq expression was violated")

    code = 2501
    name = "RequireEqViolated"
    msg = "A require_eq expression was violated"


class RequireKeysEqViolated(ProgramError):
    def __init__(self):
        super().__init__(2502, "A require_keys_eq expression was violated")

    code = 2502
    name = "RequireKeysEqViolated"
    msg = "A require_keys_eq expression was violated"


class RequireNeqViolated(ProgramError):
    def __init__(self):
        super().__init__(2503, "A require_neq expression was violated")

    code = 2503
    name = "RequireNeqViolated"
    msg = "A require_neq expression was violated"


class RequireKeysNeqViolated(ProgramError):
    def __init__(self):
        super().__init__(2504, "A require_keys_neq expression was violated")

    code = 2504
    name = "RequireKeysNeqViolated"
    msg = "A require_keys_neq expression was violated"


class RequireGtViolated(ProgramError):
    def __init__(self):
        super().__init__(2505, "A require_gt expression was violated")

    code = 2505
    name = "RequireGtViolated"
    msg = "A require_gt expression was violated"


class RequireGteViolated(ProgramError):
    def __init__(self):
        super().__init__(2506, "A require_gte expression was violated")

    code = 2506
    name = "RequireGteViolated"
    msg = "A require_gte expression was violated"


class AccountDiscriminatorAlreadySet(ProgramError):
    def __init__(self):
        super().__init__(3000, "The account discriminator was already set on this account")

    code = 3000
    name = "AccountDiscriminatorAlreadySet"
    msg = "The account discriminator was already set on this account"


class AccountDiscriminatorNotFound(ProgramError):
    def __init__(self):
        super().__init__(3001, "No 8 byte discriminator was found on the account")

    code = 3001
    name = "AccountDiscriminatorNotFound"
    msg = "No 8 byte discriminator was found on the account"


class AccountDiscriminatorMismatch(ProgramError):
    def __init__(self):
        super().__init__(3002, "8 byte discriminator did not match what was expected")

    code = 3002
    name = "AccountDiscriminatorMismatch"
    msg = "8 byte discriminator did not match what was expected"


class AccountDidNotDeserialize(ProgramError):
    def __init__(self):
        super().__init__(3003, "Failed to deserialize the account")

    code = 3003
    name = "AccountDidNotDeserialize"
    msg = "Failed to deserialize the account"


class AccountDidNotSerialize(ProgramError):
    def __init__(self):
        super().__init__(3004, "Failed to serialize the account")

    code = 3004
    name = "AccountDidNotSerialize"
    msg = "Failed to serialize the account"


class AccountNotEnoughKeys(ProgramError):
    def __init__(self):
        super().__init__(3005, "Not enough account keys given to the instruction")

    code = 3005
    name = "AccountNotEnoughKeys"
    msg = "Not enough account keys given to the instruction"


class AccountNotMutable(ProgramError):
    def __init__(self):
        super().__init__(3006, "The given account is not mutable")

    code = 3006
    name = "AccountNotMutable"
    msg = "The given account is not mutable"


class AccountOwnedByWrongProgram(ProgramError):
    def __init__(self):
        super().__init__(3007, "The given account is owned by a different program than expected")

    code = 3007
    name = "AccountOwnedByWrongProgram"
    msg = "The given account is owned by a different program than expected"


class InvalidProgramId(ProgramError):
    def __init__(self):
        super().__init__(3008, "Program ID was not as expected")

    code = 3008
    name = "InvalidProgramId"
    msg = "Program ID was not as expected"


class InvalidProgramExecutable(ProgramError):
    def __init__(self):
        super().__init__(3009, "Program account is not executable")

    code = 3009
    name = "InvalidProgramExecutable"
    msg = "Program account is not executable"


class AccountNotSigner(ProgramError):
    def __init__(self):
        super().__init__(3010, "The given account did not sign")

    code = 3010
    name = "AccountNotSigner"
    msg = "The given account did not sign"


class AccountNotSystemOwned(ProgramError):
    def __init__(self):
        super().__init__(3011, "The given account is not owned by the system program")

    code = 3011
    name = "AccountNotSystemOwned"
    msg = "The given account is not owned by the system program"


class AccountNotInitialized(ProgramError):
    def __init__(self):
        super().__init__(3012, "The program expected this account to be already initialized")

    code = 3012
    name = "AccountNotInitialized"
    msg = "The program expected this account to be already initialized"


class AccountNotProgramData(ProgramError):
    def __init__(self):
        super().__init__(3013, "The given account is not a program data account")

    code = 3013
    name = "AccountNotProgramData"
    msg = "The given account is not a program data account"


class AccountNotAssociatedTokenAccount(ProgramError):
    def __init__(self):
        super().__init__(3014, "The given account is not the associated token account")

    code = 3014
    name = "AccountNotAssociatedTokenAccount"
    msg = "The given account is not the associated token account"


class AccountSysvarMismatch(ProgramError):
    def __init__(self):
        super().__init__(3015, "The given public key does not match the required sysvar")

    code = 3015
    name = "AccountSysvarMismatch"
    msg = "The given public key does not match the required sysvar"


class StateInvalidAddress(ProgramError):
    def __init__(self):
        super().__init__(4000, "The given state account does not have the correct address")

    code = 4000
    name = "StateInvalidAddress"
    msg = "The given state account does not have the correct address"


class Deprecated(ProgramError):
    def __init__(self):
        super().__init__(5000, "The API being used is deprecated and should no longer be used")

    code = 5000
    name = "Deprecated"
    msg = "The API being used is deprecated and should no longer be used"


AnchorError = typing.Union[
    InstructionMissing,
    InstructionFallbackNotFound,
    InstructionDidNotDeserialize,
    InstructionDidNotSerialize,
    IdlInstructionStub,
    IdlInstructionInvalidProgram,
    ConstraintMut,
    ConstraintHasOne,
    ConstraintSigner,
    ConstraintRaw,
    ConstraintOwner,
    ConstraintRentExempt,
    ConstraintSeeds,
    ConstraintExecutable,
    ConstraintState,
    ConstraintAssociated,
    ConstraintAssociatedInit,
    ConstraintClose,
    ConstraintAddress,
    ConstraintZero,
    ConstraintTokenMint,
    ConstraintTokenOwner,
    ConstraintMintMintAuthority,
    ConstraintMintFreezeAuthority,
    ConstraintMintDecimals,
    ConstraintSpace,
    RequireViolated,
    RequireEqViolated,
    RequireKeysEqViolated,
    RequireNeqViolated,
    RequireKeysNeqViolated,
    RequireGtViolated,
    RequireGteViolated,
    AccountDiscriminatorAlreadySet,
    AccountDiscriminatorNotFound,
    AccountDiscriminatorMismatch,
    AccountDidNotDeserialize,
    AccountDidNotSerialize,
    AccountNotEnoughKeys,
    AccountNotMutable,
    AccountOwnedByWrongProgram,
    InvalidProgramId,
    InvalidProgramExecutable,
    AccountNotSigner,
    AccountNotSystemOwned,
    AccountNotInitialized,
    AccountNotProgramData,
    AccountNotAssociatedTokenAccount,
    AccountSysvarMismatch,
    StateInvalidAddress,
    Deprecated,
]
ANCHOR_ERROR_MAP: dict[int, AnchorError] = {
    100: InstructionMissing(),
    101: InstructionFallbackNotFound(),
    102: InstructionDidNotDeserialize(),
    103: InstructionDidNotSerialize(),
    1000: IdlInstructionStub(),
    1001: IdlInstructionInvalidProgram(),
    2000: ConstraintMut(),
    2001: ConstraintHasOne(),
    2002: ConstraintSigner(),
    2003: ConstraintRaw(),
    2004: ConstraintOwner(),
    2005: ConstraintRentExempt(),
    2006: ConstraintSeeds(),
    2007: ConstraintExecutable(),
    2008: ConstraintState(),
    2009: ConstraintAssociated(),
    2010: ConstraintAssociatedInit(),
    2011: ConstraintClose(),
    2012: ConstraintAddress(),
    2013: ConstraintZero(),
    2014: ConstraintTokenMint(),
    2015: ConstraintTokenOwner(),
    2016: ConstraintMintMintAuthority(),
    2017: ConstraintMintFreezeAuthority(),
    2018: ConstraintMintDecimals(),
    2019: ConstraintSpace(),
    2500: RequireViolated(),
    2501: RequireEqViolated(),
    2502: RequireKeysEqViolated(),
    2503: RequireNeqViolated(),
    2504: RequireKeysNeqViolated(),
    2505: RequireGtViolated(),
    2506: RequireGteViolated(),
    3000: AccountDiscriminatorAlreadySet(),
    3001: AccountDiscriminatorNotFound(),
    3002: AccountDiscriminatorMismatch(),
    3003: AccountDidNotDeserialize(),
    3004: AccountDidNotSerialize(),
    3005: AccountNotEnoughKeys(),
    3006: AccountNotMutable(),
    3007: AccountOwnedByWrongProgram(),
    3008: InvalidProgramId(),
    3009: InvalidProgramExecutable(),
    3010: AccountNotSigner(),
    3011: AccountNotSystemOwned(),
    3012: AccountNotInitialized(),
    3013: AccountNotProgramData(),
    3014: AccountNotAssociatedTokenAccount(),
    3015: AccountSysvarMismatch(),
    4000: StateInvalidAddress(),
    5000: Deprecated(),
}


def from_code(code: int) -> typing.Optional[AnchorError]:
    maybe_err = ANCHOR_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


@dataclass
class OpenOrders:
    discriminator: typing.ClassVar = b"\x73\x65\x72\x75\x6D"
    layout: typing.ClassVar = borsh.CStruct(
        "account_flags" / types.account_flags.AccountFlags.layout,
        "market" / BorshPubkey,
        "owner" / BorshPubkey,
        "base_token_free" / borsh.U64,
        "base_token_total" / borsh.U64,
        "quote_token_free" / borsh.U64,
        "quote_token_total" / borsh.U64,
        "free_slot_bits" / borsh.U128,
        "is_bid_bits" / borsh.U128,
        "orders" / borsh.U128[128],
        "client_ids" / borsh.U64[128],
        "referrer_rebate_accrued" / borsh.U64,
        "padding" / borsh.U8[7],
    )
    account_flags: types.account_flags.AccountFlags
    market: Pubkey
    owner: Pubkey
    base_token_free: int
    base_token_total: int
    quote_token_free: int
    quote_token_total: int
    free_slot_bits: int
    is_bid_bits: int
    orders: list[int]
    client_ids: list[int]
    referrer_rebate_accrued: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["OpenOrders"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["OpenOrders"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["OpenOrders"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "OpenOrders":
        if data[: len(cls.discriminator)] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = OpenOrders.layout.parse(data[len(cls.discriminator) :])
        return cls(
            account_flags=types.account_flags.AccountFlags.from_decoded(dec.account_flags),
            market=dec.market,
            owner=dec.owner,
            base_token_free=dec.base_token_free,
            base_token_total=dec.base_token_total,
            quote_token_free=dec.quote_token_free,
            quote_token_total=dec.quote_token_total,
            free_slot_bits=dec.free_slot_bits,
            is_bid_bits=dec.is_bid_bits,
            orders=dec.orders,
            client_ids=dec.client_ids,
            referrer_rebate_accrued=dec.referrer_rebate_accrued,
        )



from dataclasses import dataclass
from enum import Enum
from typing import Union

from anchorpy import Event
from construct import Container
from solders.pubkey import Pubkey

from zetamarkets_py import utils
from zetamarkets_py.types import Asset, OrderCompleteType, Side


# PlaceOrderEvent but we add the args from the PlaceOrder instruction itself, as well as the tx slot and signature
@dataclass
class PlaceOrderEventWithArgs:
    # ix args
    price: float
    size: float
    side: Side

    # ix event
    fee: float
    oracle_price: float
    order_id: int
    asset: Asset
    margin_account: Pubkey
    client_order_id: int

    @classmethod
    def from_event_and_args(cls, event: Event, args: Container):
        assert event.name.startswith("PlaceOrderEvent")
        return cls(
            price=utils.convert_fixed_int_to_decimal(args.price),
            size=utils.convert_fixed_lot_to_decimal(args.size),
            side=Side.from_index(args.side.index),
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_id=event.data.client_order_id,
        )


@dataclass
class PlaceOrderEvent:
    fee: float
    oracle_price: float
    order_id: int
    asset: Asset
    margin_account: Pubkey
    client_order_id: int

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.oracle_price),
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            margin_account=event.data.margin_account,
            client_order_id=event.data.client_order_id,
        )


@dataclass
class OrderCompleteEvent:
    margin_account: Pubkey
    authority: Pubkey
    asset: Asset
    side: Side
    unfilled_size: float
    order_id: int
    client_order_id: int
    order_complete_type: OrderCompleteType

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            margin_account=event.data.margin_account,
            authority=event.data.user,
            asset=Asset.from_index(event.data.asset.index),
            side=Side.from_index(event.data.side.index),
            unfilled_size=utils.convert_fixed_lot_to_decimal(event.data.unfilled_size),
            order_id=event.data.order_id,
            client_order_id=event.data.client_order_id,
            order_complete_type=OrderCompleteType.from_index(event.data.order_complete_type.index),
        )


@dataclass
class CancelOrderEvent:
    margin_account: Pubkey
    authority: Pubkey
    asset: Asset
    side: Side
    unfilled_size: float
    order_id: int
    client_order_id: int

    @classmethod
    def from_order_complete_event(cls, event: OrderCompleteEvent):
        return cls(
            margin_account=event.margin_account,
            authority=event.authority,
            asset=event.asset,
            side=event.side,
            unfilled_size=event.unfilled_size,
            order_id=event.order_id,
            client_order_id=event.client_order_id,
        )


@dataclass
class TradeEvent:
    margin_account: Pubkey
    price: float
    size: float
    cost_of_trades: float
    side: Side
    client_order_id: int
    order_id: int
    asset: Asset
    authority: Pubkey
    is_taker: bool
    sequence_number: int
    fee: float

    @classmethod
    def from_event(cls, event: Event):
        assert event.name.startswith(cls.__name__)
        return cls(
            margin_account=event.data.margin_account,
            price=utils.convert_fixed_int_to_decimal(event.data.cost_of_trades)
            / utils.convert_fixed_lot_to_decimal(event.data.size),
            size=utils.convert_fixed_lot_to_decimal(event.data.size),
            cost_of_trades=utils.convert_fixed_int_to_decimal(event.data.cost_of_trades),
            side=Side.Bid if event.data.is_bid else Side.Ask,
            client_order_id=event.data.client_order_id,
            order_id=event.data.order_id,
            asset=Asset.from_index(event.data.asset.index),
            authority=event.data.user,
            is_taker=event.data.is_taker,
            sequence_number=event.data.sequence_number,
            fee=utils.convert_fixed_int_to_decimal(event.data.fee),
        )


@dataclass
class LiquidationEvent:
    liquidator_reward: float
    insurance_reward: float
    side: Side
    liquidation_price: float
    liquidation_size: float
    remaining_liquidatee_balance: float
    remaining_liquidator_balance: float
    mark_price: float
    oracle_price: float
    liquidatee: Pubkey
    liquidator: Pubkey
    asset: Asset
    liquidatee_margin_account: Pubkey
    liquidator_margin_account: Pubkey

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            liquidator_reward=utils.convert_fixed_int_to_decimal(event.data.liquidator_reward),
            insurance_reward=utils.convert_fixed_int_to_decimal(event.data.insurance_reward),
            side=Side.Bid if event.data.size > 0 else Side.Ask,
            liquidation_price=utils.convert_fixed_int_to_decimal(event.data.cost_of_trades)
            / utils.convert_fixed_lot_to_decimal(abs(event.data.size)),  # TODO: check this
            liquidation_size=utils.convert_fixed_lot_to_decimal(abs(event.data.size)),
            remaining_liquidatee_balance=utils.convert_fixed_int_to_decimal(event.data.remaining_liquidatee_balance),
            remaining_liquidator_balance=utils.convert_fixed_int_to_decimal(event.data.remaining_liquidator_balance),
            mark_price=utils.convert_fixed_int_to_decimal(event.data.mark_price),
            oracle_price=utils.convert_fixed_int_to_decimal(event.data.underlying_price),
            liquidatee=event.data.liquidatee,
            liquidator=event.data.liquidator,
            asset=Asset.from_index(event.data.asset.index),
            liquidatee_margin_account=event.data.liquidatee_margin_account,
            liquidator_margin_account=event.data.liquidator_margin_account,
        )


ZetaEvent = Union[PlaceOrderEvent, TradeEvent, CancelOrderEvent, LiquidationEvent]

ZetaEnrichedEvent = Union[PlaceOrderEventWithArgs, TradeEvent, CancelOrderEvent, LiquidationEvent]


class TransactionEvent(Enum):
    """
    A place order event for the user margin account.
    """

    PlaceOrderEvent = PlaceOrderEvent
    """
    An OrderComplete event for the user margin account.
    Happens when an order is either fully filled or cancelled
    """

    OrderCompleteEvent = OrderCompleteEvent
    """
    A trade event for the user margin account.
    """
    TradeEvent = TradeEvent
    """
    A liquidation event for the user margin account.
    """
    LiquidationEvent = LiquidationEvent



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container

from . import asset


class HaltStateArgsJSON(typing.TypedDict):
    asset: asset.AssetJSON
    spot_price: int
    timestamp: int


@dataclass
class HaltStateArgs:
    layout: typing.ClassVar = borsh.CStruct("asset" / asset.layout, "spot_price" / borsh.U64, "timestamp" / borsh.U64)
    asset: asset.AssetKind
    spot_price: int
    timestamp: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "HaltStateArgs":
        return cls(
            asset=asset.from_decoded(obj.asset),
            spot_price=obj.spot_price,
            timestamp=obj.timestamp,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "asset": self.asset.to_encodable(),
            "spot_price": self.spot_price,
            "timestamp": self.timestamp,
        }

    def to_json(self) -> HaltStateArgsJSON:
        return {
            "asset": self.asset.to_json(),
            "spot_price": self.spot_price,
            "timestamp": self.timestamp,
        }

    @classmethod
    def from_json(cls, obj: HaltStateArgsJSON) -> "HaltStateArgs":
        return cls(
            asset=asset.from_json(obj["asset"]),
            spot_price=obj["spot_price"],
            timestamp=obj["timestamp"],
        )



from __future__ import annotations

import asyncio
import itertools
import logging
import time
from dataclasses import dataclass
from typing import Optional, Tuple

from solana.rpc.async_api import AsyncClient
from solders.pubkey import Pubkey

from zetamarkets_py import constants, pda, utils
from zetamarkets_py.constants import Asset
from zetamarkets_py.orderbook import Orderbook
from zetamarkets_py.serum_client.accounts.market_state import MarketState
from zetamarkets_py.serum_client.accounts.orderbook import OrderbookAccount
from zetamarkets_py.serum_client.accounts.queue import EventQueue
from zetamarkets_py.serum_client.types.queue import Event
from zetamarkets_py.types import FilledOrder, Network, Order, OrderInfo, Side


# Going to use ws for now, can add polling later
@dataclass
class Market:
    connection: AsyncClient
    zeta_program_id: Pubkey
    matching_engine_program_id: Pubkey
    asset: Asset

    _market_state: MarketState
    _base_zeta_vault_address: Pubkey
    _quote_zeta_vault_address: Pubkey
    _logger: logging.Logger
    _bids_subscription_task: Optional[asyncio.Task] = None
    _asks_subscription_task: Optional[asyncio.Task] = None
    _bids_last_update_slot: Optional[int] = None
    _asks_last_update_slot: Optional[int] = None

    @classmethod
    async def load(
        cls,
        network: Network,
        connection: AsyncClient,
        asset: Asset,
        market_state_address: Pubkey,
        log_level: int = logging.CRITICAL,
    ):
        # Initialize
        zeta_program_id = constants.ZETA_PID[network]
        matching_engine_program_id = constants.MATCHING_ENGINE_PID[network]

        # Load Market State
        _market_state = await MarketState.fetch(
            connection, market_state_address, connection.commitment, matching_engine_program_id
        )
        if _market_state is None:
            raise Exception(f"Market state not found at {market_state_address}")

        # Addresses
        _base_zeta_vault_address = pda.get_zeta_vault_address(zeta_program_id, _market_state.base_mint)
        _quote_zeta_vault_address = pda.get_zeta_vault_address(zeta_program_id, _market_state.quote_mint)

        # not currently used
        logger = utils.create_logger(f"{__name__}.{cls.__name__}.{asset.name}", log_level)

        instance = cls(
            connection=connection,
            zeta_program_id=zeta_program_id,
            matching_engine_program_id=matching_engine_program_id,
            asset=asset,
            _market_state=_market_state,
            _base_zeta_vault_address=_base_zeta_vault_address,
            _quote_zeta_vault_address=_quote_zeta_vault_address,
            _logger=logger,
        )

        return instance

    @property
    def address(self) -> Pubkey:
        return self._market_state.own_address

    @property
    def _is_subscribed_bids(self) -> bool:
        return self._bids_subscription_task is not None

    @property
    def _is_subscribed_asks(self) -> bool:
        return self._asks_subscription_task is not None

    async def print_orderbook(self, depth: int = 10) -> None:
        print("Ask Orders:")
        ask_l2 = await self.get_l2(Side.Ask, depth)
        if ask_l2 is not None:
            print(*ask_l2[::-1], sep="\n")
        print("Bid Orders:")
        bid_l2 = await self.get_l2(Side.Bid, depth)
        if bid_l2 is not None:
            print(*bid_l2, sep="\n")

    async def load_bids(self) -> Optional[Orderbook]:
        """Load the bid order book"""
        bids = await Orderbook.load(
            self.connection, self._market_state.bids, self.connection.commitment, Side.Bid, self._market_state
        )
        return bids

    async def load_asks(self) -> Optional[Orderbook]:
        """Load the ask order book."""
        asks = await Orderbook.load(
            self.connection, self._market_state.asks, self.connection.commitment, Side.Ask, self._market_state
        )
        return asks

    async def load_bids_and_asks(self) -> Tuple[Optional[Orderbook], Optional[Orderbook]]:
        """Load the bid and ask orderbooks"""
        bids_account, asks_account = await OrderbookAccount.fetch_multiple(
            self.connection,
            [self._market_state.bids, self._market_state.asks],
            self.connection.commitment,
            self.matching_engine_program_id,
        )
        bids = Orderbook(Side.Bid, bids_account, self._market_state) if bids_account else None
        asks = Orderbook(Side.Ask, asks_account, self._market_state) if asks_account else None
        return bids, asks

    async def load_orders_for_owner(self, open_orders_account_address: Pubkey) -> Optional[list[Order]]:
        """Load orders for owner."""
        bids, asks = await self.load_bids_and_asks()
        if bids is None or asks is None:
            return None
        return self._parse_orders_for_owner(bids, asks, open_orders_account_address)

    async def load_event_queue(self) -> Optional[EventQueue]:
        """Load the event queue which includes the fill item and out item. For any trades two fill items are added to
        the event queue. And in case of a trade, cancel or IOC order that missed, out items are added to the event
        queue.
        """
        eq = await EventQueue.fetch(
            self.connection, self._market_state.event_queue, self.connection.commitment, self.matching_engine_program_id
        )
        if eq is None or not (eq.header.account_flags.initialized and eq.header.account_flags.event_queue):
            raise Exception("Invalid events queue, either not initialized or not a event queue.")
        return eq

    async def load_fills(self, limit=100) -> Optional[list[FilledOrder]]:
        """Note: this method may not work since we've modified our event queue (TODO: check this))"""
        raise NotImplementedError

    async def get_l2(
        self, side: Side, depth: int = 1000, clock_ts: int = int(time.time())
    ) -> Optional[list[OrderInfo]]:
        """Get the Level 2 market information."""
        orderbook = await (self.load_bids() if side == Side.Bid else self.load_asks())
        if orderbook is None:
            return None
        return orderbook._get_l2(depth, clock_ts)

    @staticmethod
    def _parse_orders_for_owner(
        bids: Orderbook, asks: Orderbook, open_orders_account_address: Pubkey
    ) -> Optional[list[Order]]:
        all_orders = itertools.chain(bids.orders(), asks.orders())
        orders = [o for o in all_orders if str(o.open_order_address) == str(open_orders_account_address)]
        return orders

    def _parse_fills(self, events: list[Event], limit: int) -> list[FilledOrder]:
        return [
            self._parse_fill_event(event)
            for event in events
            if event.event_flags.fill and event.native_quantity_paid > 0
        ]

    def _parse_fill_event(self, event: Event) -> FilledOrder:
        if event.event_flags.bid:
            side = Side.Bid
            price_before_fees = (
                event.native_quantity_released + event.native_fee_or_rebate
                if event.event_flags.maker
                else event.native_quantity_released - event.native_fee_or_rebate
            )
        else:
            side = Side.Ask
            price_before_fees = (
                event.native_quantity_released - event.native_fee_or_rebate
                if event.event_flags.maker
                else event.native_quantity_released + event.native_fee_or_rebate
            )

        # price = (price_before_fees * self.state.base_spl_token_multiplier()) / (
        #     self.state.quote_spl_token_multiplier() * event.native_quantity_paid
        # )
        # size = event.native_quantity_paid / self.state.base_spl_token_multiplier()
        # TODO: check if this is correct
        price = utils.convert_fixed_int_to_decimal(int(price_before_fees / event.native_quantity_paid))
        size = utils.convert_fixed_lot_to_decimal(event.native_quantity_paid)
        return FilledOrder(
            order_id=event.order_id,
            side=side,
            price=price,
            size=size,
            fee_cost=event.native_fee_or_rebate * (1 if event.event_flags.maker else -1),
        )
